# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.tests.test_categories
:Synopsis:          This module is used by pytest to verify that the ``categories`` module functions properly
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     30 Sep 2022
"""

import os
import sys

import requests

from . import resources

# Define a global variable to define when the package path has been set
package_path_defined = False


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list.

    .. versionadded:: 5.1.0
    """
    global package_path_defined
    if not package_path_defined:
        sys.path.insert(0, os.path.abspath('../..'))
        package_path_defined = True


def test_get_category_id():
    """This function tests the ability to get a category ID from a URL.

    .. versionchanged:: 5.1.1
       This function has been updated to leverage the functions in the ``resources`` module.

    .. versionadded:: 5.1.0
    """
    # Get the control data and core object
    control_data, khoros_object = resources.get_control_data('categories'), resources.get_core_object()

    # Test retrieving the category ID from an example URL
    category_id = khoros_object.categories.get_category_id(control_data.get('url'))
    assert category_id == control_data.get('id')


def test_total_count():
    """This function tests the ability to retrieve the total category count.

    .. versionchanged:: 5.1.1
       This function has been updated to leverage the functions in the ``resources`` module.

    .. versionadded:: 5.1.0
    """
    # Instantiate the core object
    khoros_object = resources.get_core_object()

    # Test retrieving the total category count
    total_count = khoros_object.categories.get_total_count()
    assert isinstance(total_count, int) and total_count > 0


def test_if_category_exists():
    """This function tests if the existence of categories can be successfully determined.

    .. versionchanged:: 5.1.1
       This function has been updated to leverage the functions in the ``resources`` module.

    .. versionadded:: 5.1.0
    """
    # Get the control data and core object
    control_data, khoros_object = resources.get_control_data('categories'), resources.get_core_object()

    # Test both methods
    exists_by_id = khoros_object.categories.category_exists(category_id=control_data.get('id'))
    exists_by_url = khoros_object.categories.category_exists(category_url=control_data.get('url'))
    assert exists_by_id is True and exists_by_url is True


def test_category_details():
    """This function tests the retrieval of various category details.

    .. versionchanged:: 5.1.3
       Fixed an issue with testing the retrieval of views to ensure a successful assertion.

    .. versionchanged:: 5.1.1
       This function has been updated to leverage the functions in the ``resources`` module.

    .. versionadded:: 5.1.0
    """
    # Get the control data and core object
    control_data, khoros_object = resources.get_control_data('categories'), resources.get_core_object()

    # Test retrieval of URL
    url = khoros_object.categories.get_url(control_data.get('id'))
    assert url == control_data.get('url')

    # Test retrieval of title
    title = khoros_object.categories.get_title(control_data.get('id'))
    assert title == control_data.get('title')

    # Test retrieval of description
    desc = khoros_object.categories.get_description(control_data.get('id'))
    assert desc == control_data.get('description')

    # Test retrieval of get_parent_type()
    parent_type = khoros_object.categories.get_parent_type(control_data.get('id'))
    assert parent_type == control_data.get('parent_type')

    # Test retrieval of get_parent_id()
    parent_id = khoros_object.categories.get_parent_id(control_data.get('id'))
    assert parent_id == control_data.get('parent_id')

    # Test retrieval of get_parent_url()
    parent_url = khoros_object.categories.get_parent_url(control_data.get('id'))
    assert parent_url == control_data.get('parent_url')

    # Test retrieval of get_root_type()
    root_type = khoros_object.categories.get_root_type(control_data.get('id'))
    assert root_type == control_data.get('root_type')

    # Test retrieval of get_root_id()
    root_id = khoros_object.categories.get_root_id(control_data.get('id'))
    assert root_id == control_data.get('root_id')

    # Test retrieval of get_root_url()
    root_url = khoros_object.categories.get_root_url(control_data.get('id'))
    assert root_url == control_data.get('root_url')

    # Test retrieval of language
    language = khoros_object.categories.get_language(control_data.get('id'))
    assert language == control_data.get('language')

    # Test retrieval of hidden setting
    hidden = khoros_object.categories.is_hidden(control_data.get('id'))
    assert hidden == control_data.get('hidden')

    # Test retrieval of views
    views = khoros_object.categories.get_views(control_data.get('id'))
    assert isinstance(views, int) and views >= 0

    # Test retrieval of friendly date enabled setting
    friendly_date = khoros_object.categories.friendly_date_enabled(control_data.get('id'))
    assert friendly_date == control_data.get('friendly_date')

    # TODO: Test retrieval of friendly date max age

    # Test retrieval of active skin
    active_skin = khoros_object.categories.get_active_skin(control_data.get('id'))
    assert active_skin == control_data.get('active_skin')

    # Test retrieval of depth
    depth = khoros_object.categories.get_depth(control_data.get('id'))
    assert isinstance(depth, int)

    # Test retrieval of position
    position = khoros_object.categories.get_position(control_data.get('id'))
    assert isinstance(position, int)

    # Test retrieval of get_creation_date()
    creation_date = khoros_object.categories.get_creation_date(control_data.get('id'))
    assert creation_date == control_data.get('creation_date')


def test_create_category(monkeypatch):
    """This function tests the ability to create a category.

    .. versionadded:: 5.1.2
    """
    # Instantiate the core object
    khoros_object = resources.get_core_object()

    # Overwrite the requests.get functionality with the mock_post() function
    monkeypatch.setattr(requests, 'post', resources.mock_success_post)

    # Make the mock API call
    response = khoros_object.categories.create('test-category', 'Test Category')

    # Verify that the API call was a success
    assert response['status'] == 'success'



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.tests.test_library_import
:Synopsis:          This module is used by pytest to verify that the primary package can be imported successfully
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     09 Jun 2022
"""

import os
import sys


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    sys.path.insert(0, os.path.abspath('../..'))


# Test importing the primary khoros package
def import_pkg_operation():
    """This function imports the primary package and returns ``True`` when successful."""
    import khoros
    return True


# Verify that the overall package can be successfully imported
def test_library_import():
    """This function tests to confirm that the primary package can be imported successfully."""
    set_package_path()
    assert import_pkg_operation() is True       # nosec



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.tests.test_archives
:Synopsis:          This module is used by pytest to verify that the ``archives`` module functions properly
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     02 Nov 2022
"""

import os
import sys

import requests

from . import resources

# Define a global variable to define when the package path has been set
package_path_defined = False


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list.

    .. versionadded:: 5.1.2
    """
    global package_path_defined
    if not package_path_defined:
        sys.path.insert(0, os.path.abspath('../..'))
        package_path_defined = True


def test_archive_content(monkeypatch):
    """This function tests the ability to archive content.

    .. versionadded:: 5.1.2
    """
    # Instantiate the core object
    khoros_object = resources.get_core_object()

    # Overwrite the requests.get functionality with the mock_post() function
    monkeypatch.setattr(requests, 'post', resources.mock_success_post)

    # Perform and assess the mock API call using a message ID
    response = khoros_object.archives.archive('12345')
    assert response.get('status') == 'success'

    # Perform and assess the mock API call using a message URL
    response = khoros_object.archives.archive(message_url='https://community.example.com/t5/-/-/ta-p/12345')
    assert response.get('status') == 'success'


def test_unarchive_content(monkeypatch):
    """This function tests the ability to unarchive content.

    .. versionadded:: 5.1.2
    """
    # Instantiate the core object
    khoros_object = resources.get_core_object()

    # Overwrite the requests.get functionality with the mock_post() function
    monkeypatch.setattr(requests, 'post', resources.mock_success_post)

    # Perform and assess the mock API call using a message ID
    response = khoros_object.archives.unarchive('12345')
    assert response.get('status') == 'success'

    # Perform and assess the mock API call using a message URL
    response = khoros_object.archives.unarchive(message_url='https://community.example.com/t5/-/-/ta-p/12345')
    assert response.get('status') == 'success'


def test_archive_check():
    """This function tests the ability to check whether a message is archived.

    .. versionadded:: 5.2.0
    """
    # Instantiate the core object
    khoros_object = resources.get_core_object()

    # Define message IDs to test
    archived_content = '30127'
    online_content = '58909'

    # Perform and assess the methods
    assert khoros_object.archives.is_archived(archived_content) is True
    assert khoros_object.archives.is_archived(online_content) is False





# -*- coding: utf-8 -*-
"""
:Module:         khoros.utils.tests.test_helper_file
:Synopsis:       This module is used by pytest to verify that the helper configuration files work properly
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  09 Jun 2022
"""

import os
import sys


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    sys.path.insert(0, os.path.abspath('../..'))


def get_helper_path():
    """This function defines the appropriate path to the helper file."""
    if os.getenv('USERNAME') == 'shurtj' and sys.platform == 'win32':
        helper_path = f"{os.getenv('USERPROFILE')}\\Development\\khoros\\local\\"
    else:
        helper_path = f"{os.getenv('HOME')}/secrets/"
    return helper_path


def test_yaml_file():
    """This function tests the import of a YAML file."""
    # Import the package
    set_package_path()
    from khoros import Khoros

    # Define the full path to the helper file
    helper_path = get_helper_path()
    yaml_file = f"{helper_path}khoros_helper.yml"

    # Initialize the core object using the helper file
    khoros = Khoros(helper=yaml_file, auto_connect=False)

    # Verify that the helper configuration was imported successfully
    assert 'connection' in khoros._helper_settings      # nosec



# -*- coding: utf-8 -*-
"""
:Module:            khoros.api
:Synopsis:          This module handles interactions with the Khoros Community REST APIs
:Usage:             ``import khoros.api``
:Example:           ``json_response = khoros.api.get_request_with_retries(url, auth_dict=khoros.auth)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     14 Jun 2023
"""

import json
import os.path
import warnings
import importlib

import urllib3
import requests

from . import errors
from .utils import core_utils, log_utils

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)

# Define global variable to determine if SSL verification has been disabled
ssl_verify_disabled = None

# Define global variable to determine if suppressed message warning has been displayed
ssl_warning_shown = False


def define_headers(khoros_object=None, auth_dict=None, params=None, accept=None, content_type=None, multipart=False,
                   default_content_type=False, proxy_user_object=None):
    """This function defines the headers to use in an API call.

    .. versionchanged:: 5.3.0
       Added logging error messages when exceptions are raised.

    .. versionchanged:: 4.0.0
       Introduced the ``proxy_user_object`` parameter to allow API requests to be performed on behalf of other users.

    .. versionchanged:: 3.5.0
       An unnecessary ``else`` statement after the :py:exc:`khoros.errors.exceptions.MissingAuthDataError`
       exception is raised has been removed.

    .. versionchanged:: 2.7.5
       Added the ``default_content_type`` argument which is defined as ``False`` by default.

    .. versionchanged:: 2.7.4
       The HTTP headers were changed to be all lowercase in order to be standardized across the library.
       A function call for :py:func:`khoros.api._normalize_headers` was also introduced.

    .. versionchanged:: 2.3.0
       Added the ``multipart`` Boolean argument to remove the ``content-type`` key value pair when appropriate.

    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param params: Header parameters in a dictionary format
    :type params: dict, None
    :param accept: The ``accept`` header value (e.g. ``application/json``)
    :type accept: str, None
    :param content_type: The ``content-type`` header value (e.g. ``application/json``)
    :type content_type: str, None
    :param multipart: Defines whether the query is a ``multipart/form-data`` query (``False`` by default)
    :type multipart: bool
    :param default_content_type: Determines if ``application/json`` should be used as the default ``content-type``
                                 value if the key does not exist (``False`` by default)
    :type default_content_type: bool
    :param proxy_user_object: Instantiated :py:class:`khoros.objects.users.ImpersonatedUser` object to perform the
                              API request on behalf of a secondary user.
    :type proxy_user_object: class[khoros.objects.users.ImpersonatedUser], None
    :returns: A dictionary with the header fields and associated values
    :raises: :py:exc:`khoros.errors.exceptions.MissingAuthDataError`
    """
    if not khoros_object and not auth_dict and not proxy_user_object:
        error_msg = "The authentication data was not provided and a connection cannot be established."
        logger.error(error_msg)
        raise errors.exceptions.MissingAuthDataError()
    if proxy_user_object:
        headers = proxy_user_object.session_header
    elif auth_dict:
        headers = auth_dict['header']
    else:
        headers = khoros_object.auth['header']
    if 'content-type' not in headers and default_content_type:
        headers['content-type'] = 'application/json'
    if params:
        headers.update(params)
    if accept:
        headers['accept'] = accept
    if content_type:
        headers['content-type'] = content_type
    if multipart and 'content-type' in headers:
        del headers['content-type']
    headers = _normalize_headers(headers)
    return headers


def _normalize_headers(_headers):
    """This function normalizes the HTTP headers to ensure that the keys and values are all lowercase.

    .. versionchanged:: 2.7.5
       The function was updated to ensure that authentication/authorization tokens would not be altered.

    .. versionadded:: 2.7.4

    :param _headers: The headers dictionary
    :type  _headers: dict
    :returns: The normalized headers dictionary
    """
    _normalized_headers = {}
    _auth_keys = ['li-api-session-key', 'authorization']
    for _header_key, _header_value in _headers.items():
        if _header_key.lower() in _auth_keys:
            _normalized_headers[_header_key] = _header_value
        elif isinstance(_header_value, str):
            _normalized_headers[_header_key.lower()] = _header_value.lower()
        elif any((isinstance(_header_value, list), isinstance(_header_value, tuple))):
            _new_iterable = []
            for _inner_value in _header_value:
                _inner_value = _inner_value.lower() if isinstance(_inner_value, str) else _inner_value
                _new_iterable.append(_inner_value)
            _new_iterable = tuple(_new_iterable) if isinstance(_header_value, tuple) else _new_iterable
            _normalized_headers[_header_key.lower()] = _new_iterable
        else:
            _header_key = _header_key.lower() if isinstance(_header_key, str) else _header_key
            _normalized_headers[_header_key] = _header_value
    return _normalized_headers


def _get_json_query_string(_return_json, _include_ampersand_prefix=True):
    """This function constructs a query string for a Community API v1 query that should return JSON responses.

    :param _return_json: Indicates whether API responses should be in JSON format
    :type _return_json: bool
    :param _include_ampersand_prefix: Determines if an ampersand (``&``) prefix should be included in the string
    :type _include_ampersand_prefix: bool
    :returns: The formatted query string
    """
    _query_strings = {True: 'restapi.response_format=json', False: ''}
    _prefixes = {True: '&', False: ''}
    _query_string = ''
    if _return_json:
        _query_string = f"{_prefixes.get(_include_ampersand_prefix)}{_query_strings.get(_return_json)}"
    return _query_string


def _add_json_query_to_uri(_uri, _return_json=True):
    """This function appends the v1 query parameter to return a JSON response when appropriate.

    .. versionadded:: 4.0.0

    .. note:: No action will be taken if the URI is not recognized as a Community API v1 URI.

    :param _uri: The current URI against which the REST API call will be made
    :type _uri: str
    :param _return_json: Indicates whether the response should be in JSON format (``True`` by default)
    :type _return_json: bool
    :returns: The URI either untouched or with the added query parameter string where appropriate
    :raises: :py:exc:`TypeError`
    """
    if 'restapi/vc' in _uri and 'restapi.response_format' not in _uri and _return_json:
        _has_queries = True if '?' in _uri else False
        _uri = f"{_uri}?" if '?' not in _uri else _uri
        _uri = f"{_uri}{_get_json_query_string(_return_json, _has_queries)}"
    return _uri


def _display_ssl_verify_warning():
    """This function displays a warning if SSL verification has been disabled.

    .. versionchanged:: 5.3.0
       Added a warning log entry along with the warning message.

    .. versionchanged:: 5.0.0
       The redundant ``return`` statement has been removed.

    .. versionadded:: 4.3.0

    :returns: None
    """
    global ssl_warning_shown
    if ssl_warning_shown is False:
        # Warn that SSL warnings are being suppressed
        _warn_msg = 'SSL certificate verification has been explicitly disabled and warnings will be suppressed'
        logger.warn(_warn_msg)
        warnings.warn(_warn_msg)
        ssl_warning_shown = True

        # Suppress warnings when performing API calls without verifying SSL certificates
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def should_verify_tls(khoros_object=None):
    """This function determines whether to verify the server's TLS certificate. (``True`` by default)

    .. versionchanged:: 4.3.0
       Introduced the ``ssl_verify_disabled`` global variable to allow this check to be performed even when the
       core object is not passed to the function.

    :param khoros_object: The core Khoros object
    :type khoros_object: class[khoros.Khoros], None
    :returns: Boolean value indicating if the verification should occur
    """
    global ssl_warning_shown, ssl_verify_disabled
    verify = None
    if khoros_object is not None and 'ssl_verify' in khoros_object.core_settings:
        verify = khoros_object.core_settings.get('ssl_verify')
        if verify is False:
            _display_ssl_verify_warning()
    elif ssl_verify_disabled is True:
        verify = False
        _display_ssl_verify_warning()
    verify = True if verify is None else verify
    if ssl_verify_disabled is None:
        ssl_verify_disabled = True if verify is False else False
    return verify


def get_request_with_retries(query_url, return_json=True, khoros_object=None, auth_dict=None, headers=None,
                             verify=None, proxy_user_object=None):
    """This function performs a GET request with a total of 5 retries in case of timeouts or connection issues.

    .. versionchanged:: 4.0.0
       Introduced the ``proxy_user_object`` parameter to allow API requests to be performed on behalf of other users.
       A function call was also introduced to ensure that v1 requests that should return JSON responses are formatted
       correctly.

    .. versionchanged:: 3.4.0
       Removed an unnecessary ``pass`` statement and initially defined the ``response`` variable with a ``NoneType``
       value to prevent a linting error from being reported. Also leveraged the ``ssl_verify`` core setting.

    .. versionchanged:: 2.5.0
       Leverages new private functions and has improved error handling of JSON conversion attempts.

    :param query_url: The URI to be queried
    :type query_url: str
    :param return_json: Determines whether the response should be returned in JSON format (Default: ``True``)
    :type return_json: bool
    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param headers: Any header values (in dictionary format) to pass in the API call (optional)
    :type headers: dict, None
    :param verify: Determines whether to verify the server's TLS certificate (``True`` by default)
    :type verify: bool, None
    :param proxy_user_object: Instantiated :py:class:`khoros.objects.users.ImpersonatedUser` object to perform the
                              API request on behalf of a secondary user.
    :type proxy_user_object: class[khoros.objects.users.ImpersonatedUser], None
    :returns: The API response from the GET request (optionally in JSON format)
    :raises: :py:exc:`ValueError`, :py:exc:`TypeError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    headers = define_headers(khoros_object=khoros_object, auth_dict=auth_dict, params=headers,
                             proxy_user_object=proxy_user_object)
    verify = should_verify_tls(khoros_object) if verify is None else verify
    query_url = _add_json_query_to_uri(query_url, return_json)
    retries, response = 0, None
    while retries <= 5:
        try:
            response = requests.get(query_url, headers=headers, verify=verify)
            break
        except Exception as exc_msg:
            _report_failed_attempt(exc_msg, 'get', retries)
            retries += 1
    if retries == 6:
        _raise_exception_for_repeated_timeouts()
    return _attempt_json_conversion(response, return_json)


def _is_plaintext_payload(_headers, _payload=None):
    """This function checks to determine whether the payload for an API is in JSON or plaintext format.

    .. versionadded:: 3.1.0

    :param _headers: The headers associated with the API call
    :type _headers: dict
    :param _payload: The payload to be delivered in the API call
    :type _payload: dict, str
    :returns: Boolean value indicating whether the payload is plaintext
    :raises: :py:exc:`ValueError`
    """
    _is_plaintext = False
    if ('content-type' in _headers and _headers.get('content-type') == 'text/plain') or isinstance(_payload, str):
        _is_plaintext = True
    return _is_plaintext


def _api_request_with_payload(_url, _payload=None, _request_type='post', _headers=None, _multipart=False, _verify=None,
                              _khoros_object=None):
    """This function performs an API request while supplying a JSON payload.

    .. versionchanged:: 5.3.0
       Added logging error messages when exceptions are raised.

    .. versionchanged:: 4.3.0
       An issue has been fixed that prevented SSL verification from being disabled by the helper file setting.

    .. versionchanged:: 3.5.0
       Removed the unnecessary ``pass`` statement.

    .. versionchanged:: 3.4.0
       Support has been introduced for the ``verify`` parameter to determine if SSL certificate verification is needed.

    .. versionchanged:: 3.1.0
       The function now supports plaintext payloads.

    .. versionchanged:: 2.5.0
       The function can now be called without supplying a JSON payload.

    .. versionchanged:: 2.3.0
       Added the ability to perform multipart/form-data queries.

    :param _url: The URI to be queried
    :type _url: str
    :param _payload: The payload that accompanies the API call
    :type _payload: dict, str, None
    :param _request_type: Define the type of API call (e.g. ``get``, ``post`` or ``put``)
    :type _request_type: str
    :param _headers: Any predefined headers to be passed with the API call (optional)
    :type _headers: dict
    :param _multipart: Defines whether the query is a ``multipart/form-data`` query (``False`` by default)
    :type _multipart: bool
    :param _verify: Determines whether to verify the server's TLS certificate (``None`` by default)
    :type _verify: bool
    :param _khoros_object: The core Khoros object
    :type _khoros_object: class[khoros.Khoros], None
    :returns: The API response
    :raises: :py:exc:`khoros.errors.exceptions.InvalidRequestTypeError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`
    """
    _headers = {} if not _headers else _headers
    _verify = should_verify_tls(_khoros_object) if _verify is None else _verify
    _retries, _response = 0, None
    if not _payload:
        _response = _api_request_without_payload(_url, _request_type, _headers)
    else:
        _is_plaintext = _is_plaintext_payload(_headers, _payload)
        while _retries <= 5:
            try:
                if _request_type.lower() == "put":
                    if _multipart:
                        _response = requests.put(_url, files=_payload, headers=_headers, verify=_verify)
                    else:
                        _payload = json.dumps(_payload, default=str) if not _is_plaintext else _payload
                        _response = requests.put(_url, data=_payload, headers=_headers, verify=_verify)
                elif _request_type.lower() == "post":
                    if _multipart:
                        _response = requests.post(_url, files=_payload, headers=_headers, verify=_verify)
                    else:
                        _payload = json.dumps(_payload, default=str) if not _is_plaintext else _payload
                        _response = requests.post(_url, data=_payload, headers=_headers, verify=_verify)
                else:
                    _error_msg = 'The supplied request type for the API is not recognized.'
                    logger.error(_error_msg)
                    raise errors.exceptions.InvalidRequestTypeError()
                break
            except Exception as _exc_msg:
                _report_failed_attempt(_exc_msg, _request_type, _retries)
                _retries += 1
        if _retries == 6:
            _raise_exception_for_repeated_timeouts()
    return _response


def _api_request_without_payload(_url, _request_type, _headers, _verify=None, _khoros_object=None):
    """This function performs a ``POST`` or ``PUT`` request without an accompanying JSON payload.

    .. versionchanged:: 5.3.0
       Added logging error messages when exceptions are raised.

    .. versionchanged:: 4.3.0
       An issue has been fixed that prevented SSL verification from being disabled by the helper file setting.

    .. versionchanged:: 3.4.0
       Support has been introduced for the ``_verify`` parameter to determine if SSL certificate verification is needed.

    :param _url: The URL for the API request
    :type _url: str
    :param _request_type: The request type (e.g. ``post`` or ``put``)
    :type _request_type: str
    :param _headers: The headers associated with the API request
    :type _headers: dict
    :param _verify: Determines whether to verify the server's TLS certificate (``True`` by default)
    :type _verify: bool
    :param _khoros_object: The core Khoros object
    :type _khoros_object: class[khoros.Khoros], None
    :returns: The API response
    :raises: :py:exc:`khoros.errors.exceptions.InvalidRequestTypeError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`
    """
    _verify = should_verify_tls(_khoros_object) if _verify is None else _verify
    _retries, _response = 0, None
    while _retries <= 5:
        try:
            if _request_type.lower() == "post":
                _response = requests.post(_url, headers=_headers, verify=_verify)
            elif _request_type.lower() == "put":
                _response = requests.put(_url, headers=_headers, verify=_verify)
            else:
                _error_msg = 'The supplied request type for the API is not recognized.'
                logger.error(_error_msg)
                raise errors.exceptions.InvalidRequestTypeError()
            break
        except Exception as _exc_msg:
            _report_failed_attempt(_exc_msg, _request_type, _retries)
            _retries += 1
    if _retries == 6:
        _raise_exception_for_repeated_timeouts()
    return _response


def combine_json_and_avatar_payload(json_payload, avatar_image_path):
    """This function combines JSON payload with an uploaded avatar image (binary file) for a multipart API request.

    .. versionadded:: 2.6.0

    :param json_payload: The JSON payload for the API request
    :type json_payload: dict
    :param avatar_image_path: The full path to the avatar image to use
    :type avatar_image_path: str
    :returns: The full multipart payload for the API request
    :raises: :py:exc:`FileNotFoundError`
    """
    files_payload = format_avatar_payload(avatar_image_path)
    full_payload = {'api.request': (None, json.dumps(json_payload, default=str), 'application/json')}
    full_payload.update(files_payload)
    return full_payload


def format_avatar_payload(avatar_image_path):
    """This function structures and formats the avatar payload to be used in a multipart API request.

    .. versionadded:: 2.6.0

    :param avatar_image_path: The file path to the avatar image to use
    :type avatar_image_path: str
    :returns: The payload dictionary containing the binary file
    :raises: :py:exc:`FileNotFoundError`
    """
    return {'avatar': (f'{os.path.basename(avatar_image_path)}', open(avatar_image_path, 'rb'))}


def _report_failed_attempt(_exc_msg, _request_type, _retries):
    """This function reports a failed API call that will be retried.

    .. versionchanged:: 5.3.0
       Added error logging and changed the generic exception to a :py:exc:`RuntimeError` exception.

    .. versionchanged:: 5.0.0
       The redundant ``return`` statement has been removed.

    :param _exc_msg: The exception that was raised and captured within a try/except clause
    :param _request_type: The type of API request (e.g. ``post``, ``put`` or ``get``)
    :type _request_type: str
    :param _retries: The attempt number for the API request
    :type _retries: int
    :returns: None
    """
    _exc_name = type(_exc_msg).__name__
    if 'connect' not in _exc_name.lower():
        _error_msg = f"{_exc_name}: {_exc_msg}"
        logger.error(_error_msg)
        raise RuntimeError(_error_msg)
    _current_attempt = f"(Attempt {_retries} of 5)"
    _error_msg = f"The {_request_type.upper()} request has failed with the following exception: " + \
                 f"{_exc_name}: {_exc_msg} {_current_attempt}"
    logger.error(_error_msg)
    errors.handlers.eprint(f"{_error_msg}\n{_exc_name}: {_exc_msg}\n")


def _raise_exception_for_repeated_timeouts():
    """This function raises an exception when all API attempts (including) retries resulted in a timeout.

    .. versionchanged:: 5.3.0
       Added logging error messages when exceptions are raised.

    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.APIConnectionError`
    """
    _failure_msg = "The script was unable to complete successfully after five consecutive API timeouts. " + \
                   "Please run the script again or contact Khoros Support for further assistance."
    logger.error(_failure_msg)
    raise errors.exceptions.APIConnectionError(_failure_msg)


def payload_request_with_retries(url, request_type, json_payload=None, plaintext_payload=None, url_encoded_payload=None,
                                 return_json=True, khoros_object=None, auth_dict=None, headers=None, multipart=False,
                                 content_type=None, verify=None, proxy_user_object=None):
    """This function performs an API request that includes a payload with up to three reties as necessary.

    .. versionchanged:: 5.3.0
       Added logging error messages when exceptions are raised.

    .. versionchanged:: 4.0.0
       Introduced the ``proxy_user_object`` parameter to allow API requests to be performed on behalf of other users.

    .. versionchanged:: 3.4.0
       Support has been introduced for the ``ssl_verify`` core setting in the :py:class:`khoros.core.Khoros` object.

    .. versionchanged:: 3.2.0
       Support has been introduced for URL-encoded string payloads in POST and PUT requests.

    :param url: The URI to be queried
    :type url: str
    :param request_type: Defines the API call as a ``POST`` or ``PUT`` request
    :type request_type: str
    :param json_payload: The payload for the POST or PUT request in JSON format
    :type json_payload: dict, None
    :param plaintext_payload: The payload for the POST or PUT request in plaintext (i.e. ``text/plain``) format
    :type plaintext_payload: str, None
    :param url_encoded_payload: The payload for the POST or PUT request as a URL-encoded string
    :type url_encoded_payload: str, None
    :param return_json: Determines whether the response should be returned in JSON format (Default: ``True``)
    :type return_json: bool
    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param headers: Any header values (in dictionary format) to pass in the API call (optional)
    :type headers: dict, None
    :param multipart: Defines whether the query is a ``multipart/form-data`` query (``False`` by default)
    :type multipart: bool
    :param content_type: Allows the ``content-type`` value to be explicitly defined if necessary

                         .. note:: If this parameter is not defined then the content type will be identified based
                                   on the payload format and/or type of request.

    :type content_type: str, None
    :param verify: Determines whether to verify the server's TLS certificate (``True`` by default)
    :type verify: bool, None
    :param proxy_user_object: Instantiated :py:class:`khoros.objects.users.ImpersonatedUser` object to perform the
                              API request on behalf of a secondary user.
    :type proxy_user_object: class[khoros.objects.users.ImpersonatedUser], None
    :returns: The API response from the API request
    :raises: :py:exc:`ValueError`, :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`,
             :py:exc:`khoros.errors.exceptions.PUTRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidRequestTypeError`,
             :py:exc:`khoros.errors.exceptions.PayloadMismatchError`
    """
    # Ensure that the request type is valid
    valid_request_types = ['post', 'put']
    request_type = request_type.lower()
    if request_type not in valid_request_types:
        error_msg = 'The supplied request type for the API is not recognized.'
        logger.error(error_msg)
        raise errors.exceptions.InvalidRequestTypeError()

    # Determine if TLS certificates should be verified during API calls
    verify = should_verify_tls(khoros_object) if verify is None else verify

    # Construct the appropriate headers for the POST call
    if content_type:
        headers = define_headers(khoros_object=khoros_object, auth_dict=auth_dict, params=headers, multipart=multipart,
                                 content_type=content_type.lower(), proxy_user_object=proxy_user_object)
    elif plaintext_payload and not json_payload:
        multipart = False
        headers = define_headers(khoros_object=khoros_object, auth_dict=auth_dict, params=headers, multipart=multipart,
                                 content_type='text/plain', proxy_user_object=proxy_user_object)
    else:
        headers = define_headers(khoros_object=khoros_object, auth_dict=auth_dict, params=headers, multipart=multipart,
                                 proxy_user_object=proxy_user_object)

    # Perform the API call and retrieve the response
    if any((plaintext_payload, url_encoded_payload, json_payload)):
        if any((plaintext_payload, url_encoded_payload)) and not json_payload:
            if plaintext_payload and not url_encoded_payload:
                payload = plaintext_payload
            elif url_encoded_payload and not plaintext_payload:
                payload = url_encoded_payload
            else:
                error_msg = 'More than one payload was provided for the API call when only one is permitted.'
                logger.error(error_msg)
                raise errors.exceptions.PayloadMismatchError(request_type=request_type.upper())
        elif json_payload and not any((plaintext_payload, url_encoded_payload)):
            payload = json_payload
        else:
            error_msg = 'More than one payload was provided for the API call when only one is permitted.'
            logger.error(error_msg)
            raise errors.exceptions.PayloadMismatchError(request_type=request_type.upper())
    else:
        payload = None
    response = _api_request_with_payload(url, payload, request_type, headers, multipart, verify)
    return _attempt_json_conversion(response, return_json)


def post_request_with_retries(url, json_payload=None, plaintext_payload=None, url_encoded_payload=None,
                              return_json=True, khoros_object=None, auth_dict=None, headers=None, multipart=False,
                              content_type=None, verify=None, proxy_user_object=None):
    """This function performs a POST request with a total of 5 retries in case of timeouts or connection issues.

    .. versionchanged:: 4.0.0
       Introduced the ``proxy_user_object`` parameter to allow API requests to be performed on behalf of other users.
       A function call was also introduced to ensure that v1 requests that should return JSON responses are formatted
       correctly.

    .. versionchanged:: 3.4.0
       Support has been introduced for the ``ssl_verify`` core setting in the :py:class:`khoros.core.Khoros` object.

    .. versionchanged:: 3.2.0
       Support has been introduced for URL-encoded string payloads in POST requests.

    .. versionchanged:: 3.1.1
       The ``content_type`` parameter now gets defined as an empty string prior to calling the sub-function.

    .. versionchanged:: 3.1.0
       The function can now accept plaintext payloads and now leverages the `:py:func:payload_request_with_retries`
       function. The ``content_type`` parameter has also been introduced.

    .. versionchanged:: 2.5.0
       The function can now be called without supplying a JSON payload.

    .. versionchanged:: 2.3.0
       Added the ability to perform multipart/form-data queries.

    :param url: The URI to be queried
    :type url: str
    :param json_payload: The payload for the POST request in JSON format
    :type json_payload: dict, None
    :param plaintext_payload: The payload for the POST request in plaintext (i.e. ``text/plain``) format
    :type plaintext_payload: str, None
    :param url_encoded_payload: The payload for the POST request as a URL-encoded string
    :type url_encoded_payload: str, None
    :param return_json: Determines whether the response should be returned in JSON format (Default: ``True``)
    :type return_json: bool
    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param headers: Any header values (in dictionary format) to pass in the API call (optional)
    :type headers: dict, None
    :param multipart: Defines whether the query is a ``multipart/form-data`` query (``False`` by default)
    :type multipart: bool
    :param content_type: Allows the ``content-type`` value to be explicitly defined if necessary

                         .. note:: If this parameter is not defined then the content type will be identified based
                                   on the payload format and/or type of request.

    :type content_type: str, None
    :param verify: Determines whether to verify the server's TLS certificate (``True`` by default)
    :type verify: bool, None
    :param proxy_user_object: Instantiated :py:class:`khoros.objects.users.ImpersonatedUser` object to perform the
                              API request on behalf of a secondary user.
    :type proxy_user_object: class[khoros.objects.users.ImpersonatedUser], None
    :returns: The API response from the POST request
    :raises: :py:exc:`ValueError`, :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`,
             :py:exc:`khoros.errors.exceptions.PayloadMismatchError`
    """
    # Determine if TLS certificates should be verified during API calls
    verify = should_verify_tls(khoros_object) if verify is None else verify

    # Ensure the v1 URIs are formatted properly if the response must be in JSON format
    url = _add_json_query_to_uri(url, return_json)

    # Define the content type as necessary and perform the API call
    content_type = '' if not content_type else content_type
    return payload_request_with_retries(url, 'post', json_payload=json_payload, plaintext_payload=plaintext_payload,
                                        url_encoded_payload=url_encoded_payload, return_json=return_json,
                                        khoros_object=khoros_object, auth_dict=auth_dict, headers=headers,
                                        multipart=multipart, content_type=content_type.lower(), verify=verify,
                                        proxy_user_object=proxy_user_object)


def put_request_with_retries(url, json_payload=None, plaintext_payload=None, return_json=True, url_encoded_payload=None,
                             khoros_object=None, auth_dict=None, headers=None, multipart=False, content_type=None,
                             verify=None, proxy_user_object=None):
    """This function performs a PUT request with a total of 5 retries in case of timeouts or connection issues.

    .. versionchanged:: 4.0.0
       Introduced the ``proxy_user_object`` parameter to allow API requests to be performed on behalf of other users.
       A function call was also introduced to ensure that v1 requests that should return JSON responses are formatted
       correctly.

    .. versionchanged:: 3.4.0
       Support has been introduced for the ``ssl_verify`` core setting in the :py:class:`khoros.core.Khoros` object.

    .. versionchanged:: 3.2.0
       Support has been introduced for URL-encoded string payloads in PUT requests.

    .. versionchanged:: 3.1.1
       The ``content_type`` parameter now gets defined as an empty string prior to calling the sub-function.

    .. versionchanged:: 3.1.0
       The function can now accept plaintext payloads and now leverages the `:py:func:payload_request_with_retries`
       function. The ``content_type`` parameter has also been introduced.

    .. versionchanged:: 2.5.0
       The function can now be called without supplying a JSON payload.

    .. versionchanged:: 2.3.0
       Added the ability to perform multipart/form-data queries.

    :param url: The URI to be queried
    :type url: str
    :param json_payload: The payload for the PUT request in JSON format
    :type json_payload: dict, None
    :param plaintext_payload: The payload for the POST request in plaintext (i.e. ``text/plain``) format
    :type plaintext_payload: str, None
    :param url_encoded_payload: The payload for the POST request as a URL-encoded string
    :type url_encoded_payload: str, None
    :param return_json: Determines whether the response should be returned in JSON format (Default: ``True``)
    :type return_json: bool
    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param headers: Any header values (in dictionary format) to pass in the API call (optional)
    :type headers: dict, None
    :param multipart: Defines whether the query is a ``multipart/form-data`` query (``False`` by default)
    :type multipart: bool
    :param content_type: Allows the ``content-type`` value to be explicitly defined if necessary

                         .. note:: If this parameter is not defined then the content type will be identified based
                                   on the payload format and/or type of request.

    :type content_type: str, None
    :param verify: Determines whether to verify the server's TLS certificate (``True`` by default)
    :type verify: bool, None
    :param proxy_user_object: Instantiated :py:class:`khoros.objects.users.ImpersonatedUser` object to perform the
                              API request on behalf of a secondary user.
    :type proxy_user_object: class[khoros.objects.users.ImpersonatedUser], None
    :returns: The API response from the PUT request
    :raises: :py:exc:`ValueError`, :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.PUTRequestError`,
             :py:exc:`khoros.errors.exceptions.PayloadMismatchError`
    """
    url = _add_json_query_to_uri(url, return_json)
    content_type = '' if not content_type else content_type
    return payload_request_with_retries(url, 'put', json_payload=json_payload, plaintext_payload=plaintext_payload,
                                        url_encoded_payload=url_encoded_payload, return_json=return_json,
                                        khoros_object=khoros_object, auth_dict=auth_dict, headers=headers,
                                        multipart=multipart, content_type=content_type.lower(), verify=verify,
                                        proxy_user_object=proxy_user_object)


def _attempt_json_conversion(_response, _return_json):
    """This function attempts to convert an API response to JSON if requested.

    .. versionchanged:: 5.3.0
       Added an error log entry for the error before it is printed onscreen.

    .. versionadded:: 2.5.0

    :param _response: The API response to be converted
    :param _return_json: Indicates whether the API response should be converted
    :type _return_json: bool
    :returns: The API response that has been converted to JSON or in its original format if unable to convert
    """
    if _return_json and not isinstance(_response, dict):
        try:
            _response = _response.json()
        except Exception as _exc_msg:
            _exc_name = type(_exc_msg).__name__
            _error_msg = f"Failed to convert to JSON due to the following exception: {_exc_name}: {_exc_msg}"
            logger.error(_error_msg)
            errors.handlers.eprint(_error_msg)
    return _response


def query_successful(api_response):
    """This function reviews the API response from the Community API to verify whether the call was successful.

    :param api_response: The response from the API in JSON format
    :type api_response: dict
    :returns: Boolean indicating whether the API call was successful
    """
    try:
        success_values = ['successful', 'success']
        successful = True if api_response['status'] in success_values else False
    except (KeyError, IndexError, ValueError, TypeError):
        successful = False
    return successful


def get_results_count(api_response):
    """This function returns the number of results within a response from the Community API.

    :param api_response: The response to an API query in JSON format
    :type api_response: dict
    :returns: The number of results in the API response as an integer
    """
    return api_response['data']['size']


def get_items_list(api_response):
    """This function returns the list of ``items`` dictionaries within a response from the Community API.

    :param api_response: The response to an API query in JSON format
    :type api_response: dict
    :returns: List of ``items`` dictionaries from the API response
    """
    return api_response['data']['items']


def delete(url, return_json=False, khoros_object=None, auth_dict=None, headers=None, proxy_user_object=None,
           verify=None):
    """This function performs a DELETE request against the Core API.

    .. versionchanged:: 4.3.0
       An issue has been fixed that prevented SSL verification from being disabled by the helper file setting.

    .. versionchanged:: 4.0.0
       Introduced the ``proxy_user_object`` parameter to allow API requests to be performed on behalf of other users.

    .. versionchanged:: 3.4.0
       Support has been introduced for the ``ssl_verify`` core setting in the :py:class:`khoros.core.Khoros` object.

    :param url: The URI against which the DELETE request will be issued
    :type url: str
    :param return_json: Determines whether the response should be returned in JSON format (Default: ``False``)
    :type return_json: bool
    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param headers: Any header values (in dictionary format) to pass in the API call (optional)
    :type headers: dict, None
    :param proxy_user_object: Instantiated :py:class:`khoros.objects.users.ImpersonatedUser` object to perform the
                              API request on behalf of a secondary user.
    :type proxy_user_object: class[khoros.objects.users.ImpersonatedUser], None
    :param verify: Determines whether to verify the server's TLS certificate (``True`` by default)
    :type verify: bool, None
    :returns: The API response from the DELETE request (optionally in JSON format)
    """
    # Determine if TLS certificates should be verified during API calls
    verify = should_verify_tls(khoros_object) if verify is None else verify

    # Perform the API call to delete the asset
    headers = define_headers(khoros_object=khoros_object, auth_dict=auth_dict, params=headers,
                             proxy_user_object=proxy_user_object)
    response = requests.delete(url, headers=headers, verify=verify)
    if return_json:
        response = response.json()
    return response


def get_v1_node_collection(node_type):
    """This function retrieves the appropriate API v1 collection name for a given node type.

    .. versionchanged:: 5.3.0
       Added logging error messages when exceptions are raised.

    .. versionchanged:: 4.0.0
       Group Hubs are now supported by the function by passing the ``grouphub`` or ``group hub`` string.

    .. versionadded:: 3.5.0

    :param node_type: The node type for which to retrieve the collection (e.g. ``board``, ``category``)
    :type node_type: str
    :returns: The associated API v2 collection for the node type
    :raises: :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError`
    """
    node_collections = {
        'board': 'boards',
        'category': 'categories',
        'grouphub': 'nodes/type/key/grouphub',
        'group hub': 'nodes/type/key/grouphub',
    }
    node_collection = node_type if node_type in node_collections.values() else ''
    if node_type not in node_collections:
        error_msg = 'The node type that was provided is invalid.'
        logger.error(error_msg)
        raise errors.exceptions.InvalidNodeTypeError(val=node_type)
    return node_collections.get(node_type) if not node_collection else node_collection


def get_v1_user_path(user_id=None, user_email=None, user_login=None, user_sso_id=None, user_and_type=None,
                     trailing_slash=False):
    """This function returns the segment of an API v1 endpoint that is the path to define a user.

    .. versionchanged:: 5.3.0
       Added logging error messages when exceptions are raised.

    .. versionadded:: 4.0.0
       Introduced the ``user_and_type`` parameter that can be passed instead of a parameter for a specific type.

    .. versionadded:: 3.5.0

    :param user_id: The numeric User ID associated with a user
    :type user_id: str, int, None
    :param user_email: The email address associated with a user
    :type user_email: str, None
    :param user_login: The username (i.e. login) associated with a user
    :type user_login: str, None
    :param user_sso_id: The Single Sign-On (SSO) ID associated with a user
    :type user_sso_id: str, None
    :param user_and_type: A tuple with the first value being the user value and the second being the user type.
                          (Accepted types: ``id``, ``email``, ``login`` and ``sso_id``)
    :type user_and_type: tuple, None
    :param trailing_slash: Determines if the returned path should end with a slash (``False`` by default)
    :type trailing_slash: bool
    :returns: The API user path (e.g. ``/users/id/1234``)
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    # Parse the user_and_type parameter when present
    user_id = user_and_type[0] if user_and_type[1] == 'id' else user_id
    user_email = user_and_type[0] if user_and_type[1] == 'email' else user_email
    user_login = user_and_type[0] if user_and_type[1] == 'login' else user_login
    user_sso_id = user_and_type[0] if user_and_type[1] == 'sso_id' else user_sso_id

    # Validate the data and return the path
    if not any((user_id, user_email, user_login, user_sso_id)):
        error_msg = "A user identifier must be provided"
        logger.error(error_msg)
        raise errors.exceptions.MissingRequiredDataError(error_msg)
    slash = '/' if trailing_slash else ''
    if user_email:
        user_path = f"/users/email/{user_email}{slash}"
    elif user_login:
        user_path = f"/users/login/{user_login}{slash}"
    elif user_sso_id:
        user_path = f"/users/sso_id/{user_sso_id}{slash}"
    else:
        user_path = f"/users/id/{user_id}{slash}"
    return user_path


def perform_v1_search(khoros_object, endpoint, filter_field, filter_value, return_json=False, fail_on_no_results=False,
                      proxy_user_object=None, verify=None):
    """This function performs a search for a particular field value using a Community API v1 call.

    .. versionchanged:: 4.3.0
       An issue has been fixed that prevented SSL verification from being disabled by the helper file setting.

    .. versionchanged:: 4.0.0
       Introduced the ``proxy_user_object`` parameter to allow API requests to be performed on behalf of other users.

    .. versionchanged:: 3.5.0
       The typecheck was updated to utilize ``isinstance()`` instead of ``type()``.

    .. versionchanged:: 3.4.0
       Support has been introduced for the ``ssl_verify`` core setting in the :py:class:`khoros.core.Khoros` object.

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param endpoint: The API v1 endpoint against which to perform the search query
    :type endpoint: str
    :param filter_field: The name of the field being queried within the API v1 endpoint
    :type filter_field: str
    :param filter_value: The value associated with the field being queried
    :type filter_value: str, int
    :param return_json: Determines if the response should be returned in JSON format (``False`` by default)
    :type return_json: bool
    :param fail_on_no_results: Raises an exception if no results are returned (``False`` by default)
    :type fail_on_no_results: bool
    :param proxy_user_object: Instantiated :py:class:`khoros.objects.users.ImpersonatedUser` object to perform the
                              API request on behalf of a secondary user.
    :type proxy_user_object: class[khoros.objects.users.ImpersonatedUser], None
    :param verify: Determines whether to verify the server's TLS certificate (``True`` by default)
    :type verify: bool, None
    :returns: The API response (optionally in JSON format)
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    # Determine if TLS certificates should be verified during API calls
    verify = should_verify_tls(khoros_object) if verify is None else verify

    # Prepare the API call
    headers = define_headers(khoros_object, content_type='application/x-www-form-urlencoded',
                             proxy_user_object=proxy_user_object)
    if isinstance(filter_value, str):
        filter_value = core_utils.url_encode(filter_value)
    uri = f"{khoros_object.core['v1_base']}/search/{endpoint}?q={filter_field}:{filter_value}"
    uri = f"{uri}{_get_json_query_string(return_json)}"

    # Perform the API call
    response = requests.get(uri, headers=headers, verify=verify)
    if return_json:
        response = response.json()
        response = response['response'] if 'response' in response else response
    errors.handlers.verify_v1_response(response, 'get', 'users', fail_on_no_results)
    return response


def encode_multipart_data(data_fields):
    """This function uses the Streaming Multipart Data Encoder to encode the payload for a multipart/form-data API call.

    .. versionchanged:: 3.5.0
       This function now import the ``requests_toolbelt`` module locally to avoid dependency errors during installation.

    .. versionadded:: 2.3.0

    .. seealso:: This function follows the `requests_toolbelt <https://rsa.im/3dg7QiZ>`_ documentation.

    :param data_fields: A dictionary with the data to be encoded
    :type data_fields: dict
    :returns: The encoded data for use by the :py:mod:`requests` library
    """
    try:
        toolbelt_encoder = importlib.import_module('requests_toolbelt.multipart.encoder')
        multipart_data = toolbelt_encoder.MultipartEncoder(fields=data_fields)
    except ModuleNotFoundError:
        error_msg = "The 'requests_toolbelt' module is required to encode multipart data. " \
                    "Install with pip and try again."
        logger.error(error_msg)
        raise ModuleNotFoundError(error_msg)
    return multipart_data


def encode_payload_values(payload_dict):
    """This function URL-encoded any string-formatted payload values.

    .. versionadded:: 3.2.0

    :param payload_dict: The JSON payload for an API call in dictionary format
    :type payload_dict: dict
    :returns: The JSON payload with URL-encoded string values
    """
    encoded_payload = {}
    for field, value in payload_dict.items():
        if isinstance(value, str):
            encoded_payload[field] = core_utils.url_encode(value)
        else:
            encoded_payload[field] = value
    return encoded_payload


def encode_v1_query_string(query_dict, return_json=True, json_payload=False):
    """This function formats and URL-encodes a Community API v1 query string.

    .. versionchanged:: 3.2.0
       Introduced the ability to pass the query parameters as JSON payload to avoid URI length limits.

    .. versionadded:: 2.5.0

    :param query_dict: A dictionary with the query fields and associated values
    :type query_dict: dict
    :param return_json: Determines if JSON should be returned rather than XML (default: ``True``)
    :type return_json: bool
    :param json_payload: Determines if query parameters should be passed as JSON payload rather than in the URI
                         (``False`` by default)
    :type json_payload: bool
    :returns: The properly formatted and encoded query string
    """
    if return_json:
        query_dict['restapi.response_format'] = 'json'
    return core_utils.encode_query_string(query_dict, json_payload=json_payload)


def make_v1_request(khoros_object, endpoint, query_params=None, request_type='GET', return_json=True,
                    params_in_uri=False, json_payload=False, proxy_user_object=None, verify=None):
    """This function makes a Community API v1 request.

    .. versionchanged:: 5.3.0
       Added logging error messages when exceptions are raised.

    .. versionchanged:: 4.3.0
       An issue has been fixed that prevented SSL verification from being disabled by the helper file setting.

    .. versionchanged:: 4.0.0
       Introduced the ``proxy_user_object`` parameter to allow API requests to be performed on behalf of other users.

    .. versionchanged:: 3.2.0
       Introduced the new default ability to pass the query parameters as payload to avoid URI length limits,
       and fixed an issue with GET requests not returning JSON responses even when requested.

    .. versionchanged:: 3.0.0
       The ``query_params`` argument has been updated to be optional and a full query string can now
       be passed within the ``endpoint`` argument.

    .. versionchanged:: 2.7.4
       The HTTP headers were changed to be all lowercase in order to be standardized across the library.

    .. versionchanged:: 2.7.1
       Fixed a syntax error in raising the the :py:exc:`khoros.errors.exceptions.CurrentlyUnsupportedError`
       exception class and removed unnecessary print debugging.

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param endpoint: The API endpoint to be queried
    :type endpoint: str
    :param query_params: The field and associated values to be leveraged in the query string
    :type query_params: dict, None
    :param request_type: Determines which type of API request to perform (e.g. ``GET`` or ``POST``)

                         .. caution:: While ``PUT`` requests are technically supported in this library, at this time
                                      they are not yet supported by the Khoros Community API v1 endpoints.

    :type request_type: str
    :param return_json: Determines if the response should be returned in JSON format rather than the default
    :type return_json: bool
    :param params_in_uri: Determines if query parameters should be passed in the URI rather than in the request body
                         (``False`` by default)
    :type params_in_uri: bool
    :param json_payload: Determines if query parameters should be passed as JSON payload rather than in the URI
                         (``False`` by default)

                         .. caution:: This is not yet fully supported and therefore should not be used at this time.

    :type json_payload: bool
    :param proxy_user_object: Instantiated :py:class:`khoros.objects.users.ImpersonatedUser` object to perform the
                              API request on behalf of a secondary user.
    :type proxy_user_object: class[khoros.objects.users.ImpersonatedUser], None
    :param verify: Determines whether to verify the server's TLS certificate (``True`` by default)
    :type verify: bool, None
    :returns: The API response
    :raises: :py:exc:`ValueError`, :py:exc:`TypeError`,
             :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`,
             :py:exc:`khoros.errors.exceptions.PUTRequestError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.CurrentlyUnsupportedError`,
             :py:exc:`khoros.errors.exceptions.InvalidRequestTypeError`,
             :py:exc:`khoros.errors.exceptions.PayloadMismatchError`
    """
    # Determine if TLS certificates should be verified during API calls
    verify = should_verify_tls(khoros_object) if verify is None else verify

    # Construct the API request
    currently_unsupported_types = ['UPDATE', 'PATCH', 'DELETE']
    query_params = {} if not query_params else query_params
    query_string = encode_v1_query_string(query_params, return_json, params_in_uri)
    header = {"content-type": "application/x-www-form-urlencoded"}
    endpoint = endpoint[1:] if endpoint.startswith('/') else endpoint
    url = f"{khoros_object.core['v1_base']}/{endpoint}"

    # Only add the query parameters to the URI when explicitly requested
    if params_in_uri:
        query_string_delimiter = '&' if '?' in url else '?'
        url = f"{url}{query_string_delimiter}{query_string}"

    # Add query string to GET request URIs when a JSON response has been requested
    if request_type.upper() == 'GET' and return_json:
        json_query_string = 'restapi.response_format=json'
        query_string_delimiter = '&' if '?' in url else '?'
        url = f"{url}{query_string_delimiter}{json_query_string}"

    # Determine the request type and perform the appropriate call
    if request_type.upper() == 'GET':
        response = get_request_with_retries(url, return_json, khoros_object, headers=header,
                                            proxy_user_object=proxy_user_object, verify=verify)
    elif request_type.upper() == 'POST':
        if params_in_uri:
            response = post_request_with_retries(url, return_json=return_json, khoros_object=khoros_object,
                                                 headers=header, proxy_user_object=proxy_user_object, verify=verify)
        elif json_payload:
            # TODO: Finish testing and adding support for this type of payload (may just need different header)
            response = post_request_with_retries(url, json_payload=query_params, return_json=return_json,
                                                 khoros_object=khoros_object, headers=header,
                                                 proxy_user_object=proxy_user_object, verify=verify)
        else:
            response = post_request_with_retries(url, url_encoded_payload=query_string, return_json=return_json,
                                                 khoros_object=khoros_object, headers=header,
                                                 proxy_user_object=proxy_user_object, verify=verify)
    elif request_type.upper() == 'PUT':
        if params_in_uri:
            response = put_request_with_retries(url, return_json=return_json, khoros_object=khoros_object,
                                                headers=header, proxy_user_object=proxy_user_object, verify=verify)
        elif json_payload:
            # TODO: Finish testing and adding support for this type of payload (may just need different header)
            response = put_request_with_retries(url, json_payload=query_params, return_json=return_json,
                                                khoros_object=khoros_object, headers=header,
                                                proxy_user_object=proxy_user_object, verify=verify)
        else:
            response = put_request_with_retries(url, url_encoded_payload=query_string, return_json=return_json,
                                                khoros_object=khoros_object, headers=header,
                                                proxy_user_object=proxy_user_object, verify=verify)
    elif request_type.upper() in currently_unsupported_types:
        # TODO: Pass the request type into the exception to provide a more specific error
        error_msg = 'This feature is currently unsupported at this time.'
        logger.error(error_msg)
        raise errors.exceptions.CurrentlyUnsupportedError()
    else:
        # TODO: Pass the request type into the exception to provide a more specific error
        error_msg = 'This feature is currently unsupported at this time.'
        logger.error(error_msg)
        raise errors.exceptions.InvalidRequestTypeError()
    return response


def deliver_v2_results(response, full_response=None, return_id=None, return_url=None, return_api_url=None,
                       return_http_code=None, return_status=None, return_error_messages=None, split_errors=False,
                       khoros_object=None):
    """This function parses a Community API v2 response and returned specific data based on the function arguments.

    .. versionchanged:: 2.8.0
       Introduced the ability for error messages to be translated where possible to be more relevant, and added
       the optional ``khoros_object`` argument to facilitate this.

    .. versionchanged:: 2.5.2
       Replaced the ``return_developer_message`` argument with ``return_error_messages``.

    .. versionadded:: 2.5.0
       The code for this function was extracted from the :py:func:`khoros.objects.messages.create` function.

    :param response: The API response to be parsed
    :param full_response: Determines if the full raw API response should be returned
    :type full_response: bool, None
    :param return_id: Determines if the ``id`` field value should be returned
    :type return_id: bool, None
    :param return_url: Determines if the ``view_href`` field value should be returned
    :type return_url: bool, None
    :param return_api_url: Determines if the ``href`` field value should be returned
    :type return_api_url: bool, None
    :param return_http_code: Determines if the ``http_code`` field value should be returned
    :type return_http_code: bool, None
    :param return_status: Determines if the ``http_code`` field value should be returned
    :type return_status: bool, None
    :param return_error_messages: Determines if error messages should be returned when applicable
    :type return_error_messages: bool, None
    :param split_errors: Determines if error messages should be split into separate values or merged when applicable
    :type split_errors: bool
    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros], None

                         .. note:: The core object is only leveraged to check whether the ``translate_errors``
                                   setting is configured and to retrieve its value where possible.

    :returns: Boolean value indicating a successful outcome (default), the full API response or one or more specific
              fields defined by function arguments
    """
    outcome = query_successful(response)
    if any((return_id, return_url, return_api_url, return_http_code, return_status, return_error_messages)):
        data_to_return = []
        return_booleans = {
            'return_id': return_id,
            'return_url': return_url,
            'return_http_code': return_http_code,
            'return_api_url': return_api_url,
            'return_status': return_status,
            'return_error_messages': return_error_messages,
        }
        return_values = _get_v2_return_values(return_booleans, response, split_errors, khoros_object)
        for return_key, return_value in return_booleans.items():
            if return_value:
                data_to_return.append(return_values.get(return_key))
        outcome = tuple(data_to_return)
        if len(data_to_return) == 1:
            outcome = outcome[0]
    return response if full_response else outcome


def parse_v2_response(json_response, return_dict=False, status=False, error_msg=False, dev_msg=False,
                      split_errors=False, http_code=False, data_id=False, data_url=False,
                      data_api_uri=False, v2_base='', khoros_object=None):
    """This function parses an API response for a Community API v2 operation and returns parsed data.

    .. versionchanged:: 2.8.0
       Introduced the ability for error messages to be translated where possible to be more relevant, and added
       the optional ``khoros_object`` argument to facilitate this.

    .. versionchanged:: 2.5.2
       Replaced the ``developer_msg`` argument with ``error_msg`` and added the ``dev_msg`` and ``split_error``
       arguments with their accompanying functionality. (See :doc:`changelog` for more details.)

    .. versionchanged:: 2.5.0
       Moved from the :py:mod:`khoros.objects.messages` module to :py:mod:`khoros.api` and expanded the scope to
       apply to most Community API v2 API responses.

    .. versionadded:: 2.3.0

    :param json_response: The API response in JSON format
    :type json_response: dict
    :param return_dict: Defines if the parsed data should be returned within a dictionary
    :type return_dict: bool
    :param status: Defines if the **status** value should be returned
    :type status: bool
    :param error_msg: Defines if any **error messages** should be returned when applicable
    :type error_msg: bool
    :param dev_msg: Defines if the **developer message** should be returned when applicable
    :type dev_msg: bool
    :param split_errors: Defines if error messages should be returned as separate values when applicable
    :type split_errors: bool
    :param http_code: Defines if the **HTTP status code** should be returned
    :type http_code: bool
    :param data_id: Defines if the **ID** should be returned
    :type data_id: bool
    :param data_url: Defines if the **URL** should be returned
    :type data_url: bool
    :param data_api_uri: Defines if the **API URI** should be returned
    :type data_api_uri: bool
    :param v2_base: The base URL for the API v2
    :type v2_base: str, None
    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros], None

                         .. note:: The core object is only leveraged to check whether the ``translate_errors``
                                   setting is configured and to retrieve its value where possible.

    :returns: A string, tuple or dictionary with the parsed data
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    parsed_data = {}
    dev_msg = True if error_msg or dev_msg else False
    fields = {
        'status': (status, ('status',)),
        'error_msg': (error_msg, ('message',)),
        'dev_msg': (dev_msg, ('data', 'developer_message')),
        'http_code': (http_code, ('http_code',)),
        'data_id': (data_id, ('data', 'id')),
        'data_url': (data_url, ('data', 'view_href')),
        'data_api_uri': (data_api_uri, ('data', 'href')),
    }
    _confirm_field_supplied(fields)
    for field, info in fields.items():
        requested, json_path = info[0], info[1]
        if requested:
            if len(json_path) == 1:
                value = json_response[json_path[0]]
            else:
                value = json_response[json_path[0]][json_path[1]]
            parsed_data[field] = value
    if 'data_api_uri' in parsed_data and v2_base != '':
        parsed_data['data_api_uri'] = f"{v2_base}/{parsed_data.get('data_api_uri')}"
    if 'http_code' in parsed_data:
        try:
            parsed_data['http_code'] = int(parsed_data.get('http_code'))
        except (TypeError, ValueError):
            pass
    if 'error_msg' in parsed_data:
        translated_error_msg = errors.translations.translate_error(parsed_data.get('error_msg'), khoros_object)
        translated_dev_msg = errors.translations.translate_error(parsed_data.get('dev_msg'), khoros_object)
        if split_errors:
            parsed_data['error_msg'] = (translated_error_msg, translated_dev_msg)
        else:
            if len(parsed_data.get('dev_msg')) > 0 and parsed_data.get('error_msg') != parsed_data.get('dev_msg'):
                parsed_data['error_msg'] = f"{translated_error_msg} - {translated_dev_msg}"
            else:
                parsed_data['error_msg'] = translated_error_msg
        del parsed_data['dev_msg']
    if not return_dict:
        parsed_data = tuple(list(parsed_data.values()))
        if len(parsed_data) == 1:
            parsed_data = parsed_data[0]
    return parsed_data


def _get_v2_return_values(_return_booleans, _api_response, _split_errors, _khoros_object=None):
    """This function collects the relevant return values to be delivered for certain functions.

    .. versionchanged:: 2.8.0
       Introduced the ability for error messages to be translated where possible to be more relevant, and added
       the optional ``_khoros_object`` argument to facilitate this.

    .. versionadded:: 2.5.2

    :param _return_booleans: Dictionary of the return types with their associated Boolean value
    :type _return_booleans: dict
    :param _api_response: The Khoros Community API v2 response
    :type _api_response: dict
    :param _split_errors: Defines whether error messages should be merged when applicable
    :type _split_errors: bool
    :param _khoros_object: The core :py:class:`khoros.Khoros` object
    :type _khoros_object: class[khoros.Khoros], None

                          .. note:: The core object is only leveraged to check whether the ``translate_errors``
                                    setting is configured and to retrieve its value where possible.

    :returns: A dictionary of the parsed return values for return types whose Boolean values are ``True``
    """
    _return_values = {}
    for _return_type, _return_boolean in _return_booleans.items():
        if _return_boolean:
            try:
                if _return_type == 'return_id':
                    _return_values[_return_type] = parse_v2_response(_api_response, data_id=True,
                                                                     khoros_object=_khoros_object)
                elif _return_type == 'return_url':
                    _return_values[_return_type] = parse_v2_response(_api_response, data_url=True,
                                                                     khoros_object=_khoros_object)
                elif _return_type == 'return_api_url':
                    _return_values[_return_type] = parse_v2_response(_api_response, data_api_uri=True,
                                                                     khoros_object=_khoros_object)
                elif _return_type == 'return_http_code':
                    _return_values[_return_type] = parse_v2_response(_api_response, http_code=True,
                                                                     khoros_object=_khoros_object)
                elif _return_type == 'return_status':
                    _return_values[_return_type] = parse_v2_response(_api_response, status=True,
                                                                     khoros_object=_khoros_object)
                else:
                    _return_values[_return_type] = parse_v2_response(_api_response, error_msg=True,
                                                                     split_errors=_split_errors,
                                                                     khoros_object=_khoros_object)
            except KeyError:
                pass
    return _return_values


def _confirm_field_supplied(_fields_dict):
    """This function checks to ensure that at least one field has been enabled to retrieve.

    .. versionchanged:: 5.3.0
       Added logging error messages when exceptions are raised.

    .. versionchanged:: 5.0.0
       The redundant ``return`` statement has been removed.

    .. versionchanged:: 2.5.0
       Moved from the :py:mod:`khoros.objects.messages` module to :py:mod:`khoros.api`.

    .. versionadded:: 2.3.0

    :param _fields_dict: A dictionary made up of API fields and corresponding values
    :type _fields_dict: dict
    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    _field_supplied = False
    for _field_value in _fields_dict.values():
        if _field_value[0]:
            _field_supplied = True
            break
    if not _field_supplied:
        _error_msg = "At least one field must be enabled to retrieve a response."
        logger.error(_error_msg)
        raise errors.exceptions.MissingRequiredDataError(_error_msg)


def _normalize_base_url(_base_url):
    """This function normalizes the base URL (i.e. top-level domain) for use in other functions.

    .. versionadded:: 3.0.0

    :param _base_url: The base URL of a Khoros Community environment
    :type _base_url: str
    :returns: The normalized base URL
    """
    _base_url = _base_url[:-1] if _base_url.endswith('/') else _base_url
    _base_url = f"https://{_base_url}" if not _base_url.startswith('http') else _base_url
    return _base_url


def get_platform_version(base_url, full_release=False, simple=False, commit_id=False, timestamp=False,
                         khoros_object=None, verify=None):
    """This function retrieves the Khoros Community platform version information for a given environment.

    .. versionchanged:: 4.3.0
       An issue has been fixed that prevented SSL verification from being disabled by the helper file setting.

    .. versionchanged:: 3.4.0
       Support has been introduced for the ``ssl_verify`` core setting in the :py:class:`khoros.core.Khoros` object.

    .. versionadded:: 3.0.0

    :param base_url: The base URL (i.e. top-level domain) of the Khoros Community environment
    :type base_url: str
    :param full_release: Defines if the full platform release version should be returned

                         .. note:: If none of the options are enabled then the ``full_release`` option will be
                                   enabled by default.

    :type full_release: bool
    :param simple: Defines if the simple X.Y version (e.g. 20.6) should be returned
    :type simple: bool
    :param commit_id: Defines if the Commit ID (i.e. hash) for the release should be returned
    :type commit_id: bool
    :param timestamp: Defines if the timestamp of the release (e.g. 2007092156) should be returned
    :type timestamp: bool
    :param khoros_object: The core Khoros object (Optional unless needing to determine SSL certificate verification)
    :type khoros_object: class[khoros.Khoros], None
    :param verify: Determines whether to verify the server's TLS certificate (``True`` by default)
    :type verify: bool, None
    :returns: One or more string with version information
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    # Determine if TLS certificates should be verified during API calls
    verify = should_verify_tls(khoros_object) if verify is None else verify

    # Perform and parse the
    full_release = True if not any((full_release, simple, commit_id, timestamp)) else full_release
    base_url = _normalize_base_url(base_url)
    version_info = requests.get(f'{base_url}/status/version', verify=verify)
    if version_info.status_code != 200:
        fail_msg = f'The attempt to get the platform version failed with a {version_info.status_code} status code.'
        logger.error(fail_msg)
        raise errors.exceptions.GETRequestError(fail_msg)
    versions = []
    parsed_info = {
        version_info.text.split('(')[1].split(')')[0]: full_release,
        version_info.text.split('Revision: ')[1].split(' (')[0]: simple,
        version_info.text.split('Commit Id: ')[1].split(' <br>')[0]: commit_id,
        version_info.text.split('Timestamp: ')[1].split('<')[0]: timestamp
    }
    for parsed_value, enabled in parsed_info.items():
        if enabled:
            versions.append(parsed_value)
    return versions[0] if len(versions) == 1 else tuple(versions)



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects.labels
:Synopsis:          This module includes functions that handle labels within a Khoros Community environment
:Usage:             ``from khoros.objects import labels``
:Example:           ``labels = khoros.labels.get_labels_for_message(12345)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     06 Jul 2023
"""

from .. import api, liql, errors
from ..utils import log_utils

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)


def get_labels_for_message(khoros_object, message_id):
    """This function retrieves the labels associated with a specific message.

    .. versionadded:: 5.3.0

    :param khoros_object: The Khoros object initialized via the :py:mod:`khoros.core` module
    :type khoros_object: class[khoros.Khoros]
    :param message_id: The ID associated with the message to query
    :type message_id: str, int
    :returns: A list of strings for the labels
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    # Query for the labels
    query = f"SELECT * FROM labels WHERE messages.id = '{message_id}'"
    response = liql.perform_query(khoros_object, liql_query=query)

    # Identify the labels and store in a list
    message_labels = []
    for item in response['data']['items']:
        msg_label = item.get('text')
        message_labels.append(msg_label)

    # Return the compiled list
    return message_labels



# -*- coding: utf-8 -*-
"""
:Module:            khoros.structures
:Synopsis:          This module contains submodules that are specific to the various Community structures.
:Usage:             ``from khoros import structures``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     20 Sep 2022
"""

__all__ = ['base', 'boards', 'categories', 'communities', 'grouphubs', 'nodes']

# Import all submodules by default
from . import boards
from . import categories
from . import communities
from . import grouphubs
from . import nodes



# -*- coding: utf-8 -*-
"""
:Module:            khoros.errors.handlers
:Synopsis:          Functions that handle various error situations within the namespace
:Usage:             ``from khoros.errors import handlers``
:Example:           ``error_msg = handlers.get_error_from_html(html_string)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     20 Sep 2022
"""

import re
import sys
import importlib

from ..utils import core_utils


def eprint(*args, **kwargs):
    """This function behaves the same as the ``print()`` function but is leveraged to print errors to ``sys.stderr``.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    print(*args, file=sys.stderr, **kwargs)


def _exceptions_module_imported():
    """This function checks to see whether the ``exceptions`` global variable is defined."""
    try:
        exceptions
        _module_found = True
    except NameError:
        _module_found = False
    return _module_found


def _import_exceptions_module():
    """This function imports :py:func:`khoros.errors.exceptions` as a global variable using :py:mod:`importlib`.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    global exceptions
    exceptions = importlib.import_module('khoros.errors.exceptions')


def _import_exception_classes():
    """This function imports the :py:func:`khoros.errors.exceptions` module if not already imported.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    if not _exceptions_module_imported():
        _import_exceptions_module()


def get_error_from_html(html_error, v1=False):
    """This function parses an error message from Khoros displayed in HTML format.

    .. versionchanged:: 2.0.0
       Added the ``v1`` Boolean argument

    :param html_error: The raw HTML returned via the :py:mod:`requests` module
    :type html_error: str
    :param v1: Determines if the error is from a Community API v1 call (``False`` by default)
    :type v1: bool
    :returns: The concise error message parsed from the HTML
    """
    error_title = re.sub(r'</h1>.*$', r'', re.sub(r'^.*<body><h1>', r'', html_error))
    error_description = re.sub(r'</u>.*$', r'', re.sub(r'^.*description</b>\s*<u>', r'', html_error))
    error_msg = f"{error_title}{error_description}"
    if v1:
        error_msg = f"The Community API v1 call failed with the following error:\n\t{error_msg}"
    return error_msg


def get_error_from_xml(xml_error, endpoint='', fail_on_no_results=True):
    """This function retrieves any API errors returned from one of the Community APIs in XML format.

    :param xml_error: The API response in JSON format
    :type xml_error: str
    :param endpoint: The endpoint being queried by the API call (optional)
    :type endpoint: str
    :param fail_on_no_results: Defines if an exception should be raised if no results are returned (``True`` by default)
    :type fail_on_no_results: bool
    :returns: A Boolean value stating if an error was found (optional) and the error details in a tuple
    """
    error_code, error_msg, exc_msg = 0, '', ''
    no_results_pattern = re.compile(r'<response status="success">\s{2,}<\w*/></response>')
    if endpoint:
        if f'<response status="success">\n  <{endpoint}/>\n</response>' in xml_error:
            error_found = fail_on_no_results
            error_code = 404
            error_msg = f"The Community API v1 call against the '{endpoint}' endpoint returned no results."
            exc_msg = error_msg
        else:
            error_found = False
    elif no_results_pattern.search(xml_error.replace('\n', '')):
        error_found = fail_on_no_results
        error_code = 404
        error_msg = f"The Community API v1 call returned no results."
        exc_msg = error_msg
    elif '<response status="error">' in xml_error:
        error_found = True
        try:
            error_code = int(re.sub(r'.*code="', '', xml_error.replace('\n', '')).split('"')[0])
        except (ValueError, TypeError):
            error_code = 0
        error_msg = re.sub(r'\s{2,}</message>.*', '',
                           re.sub(r'.*<message>\s{2,}', '', xml_error.replace('\n', '')))
        error_msg = core_utils.decode_html_entities(error_msg)
        exc_msg = f"The Community API v1 call failed with the error code '{error_code}' and the following " + \
                  f"message: {error_msg}"
    else:
        error_found = False
    return error_found, (error_code, error_msg, exc_msg)


def _get_v1_error_from_json(_json_error, _fail_on_no_results):
    """This function extracts error details (if present) from a Community API v1 response.

    :param _json_error: The API response in JSON format
    :type _json_error: dict
    :param _fail_on_no_results: Defines if an exception should be raised if no results are returned
    :type _fail_on_no_results: bool
    :returns: A Boolean stating if an error was found and a tuple with the error code and error/exception messages
    """
    _error_code, _error_msg, _exc_msg = 0, '', ''
    _error_found = True if _json_error['status'] == 'error' else False
    if _error_found:
        _error_code = _json_error['error']['code']
        _error_msg = _json_error['error']['message']
        _exc_msg = f"The Community API v1 call failed with the error code '{_error_code}' and the following " + \
                   f"message: {_error_msg}"
    _inner_response = _json_error[list(_json_error.keys())[1]]
    if len(_inner_response[list(_inner_response.keys())[0]]) == 0:
        _error_found = _fail_on_no_results
        _error_code = 404
        _error_msg = f"The Community API v1 call returned no results."
        _exc_msg = _error_msg
    return _error_found, (_error_code, _error_msg, _exc_msg)


def _get_v2_error_from_json(_json_error):
    """This function extracts error details (if present) from a Community API v2 response.

    :param _json_error: The API response in JSON format
    :type _json_error: dict
    :returns: A Boolean stating if an error was found and a tuple with the error code and error/exception messages
    """
    _error_code, _error_msg, _exc_msg = 0, '', ''
    _error_found = True if _json_error['status'] == 'error' else False
    if _error_found:
        _error_code = _json_error['data']['code']
        _error_type = _json_error['data']['type']
        _error_msg = f"{_json_error['message']} (Error Type: {_error_type})"
        _exc_msg = f"The Community API v2 call failed with the error code '{_error_code}' and the following " + \
                   f"message: {_error_msg}"
    return _error_found, (_error_code, _error_msg, _exc_msg)


def get_error_from_json(json_error, v1=False, include_error_bool=True, fail_on_no_results=True):
    """This function retrieves any API errors returned from one of the Community APIs in JSON format.

    :param json_error: The API response in JSON format
    :type json_error: dict
    :param v1: Determines if the error is from a Community API v1 call (``False`` by default)
    :type v1: bool
    :param include_error_bool: Returns a Boolean as well that defines if an error was found (``True`` by default)
    :type include_error_bool: bool
    :param fail_on_no_results: Defines if an exception should be raised if no results are returned (``True`` by default)
    :type fail_on_no_results: bool
    :returns: A Boolean value stating if an error was found (optional) and the error details in a tuple
    """
    if 'response' in json_error or 'error' in json_error:
        v1 = True
        if 'response' in json_error:
            json_error = json_error['response']
    if v1:
        error_found, error_details = _get_v1_error_from_json(json_error, fail_on_no_results)
    else:
        # TODO: Handle v2 responses with no results similar to what is being done with v1
        error_found, error_details = _get_v2_error_from_json(json_error)
    if include_error_bool:
        return error_found, error_details
    return error_details


def verify_core_object_present(khoros_object):
    """This function verifies whether the core object was supposed and raises an exception if not.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if not khoros_object:
        _import_exception_classes()
        raise exceptions.MissingRequiredDataError('The core object must be provided in order to perform the action')


def verify_v1_response(api_response, query_type='get', endpoint='', fail_on_no_results=False):
    """This function evaluates a Community API v1 response to identify any failures.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    :param api_response: The response from the API call
    :param query_type: The type of API call that was made, such as ``get`` (default), ``post``, ``put``, etc.
    :type query_type: str
    :param endpoint: The endpoint being queried by the API call (optional)
    :type endpoint: str
    :param fail_on_no_results: Raises an exception if no results are returned (``False`` by default)
    :type fail_on_no_results: bool
    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.APIRequestError`, :py:exc:`khoros.errors.exceptions.DELETERequestError`,
             :py:exc:`khoros.errors.exceptions.GETRequestError`, :py:exc:`khoros.errors.exceptions.POSTRequestError`,
             :py:exc:`khoros.errors.exceptions.PUTRequestError`
    """
    _import_exception_classes()
    valid_query_types = ['get', 'post', 'put', 'delete']
    query_type = 'other' if query_type not in valid_query_types else query_type
    exception_classes = {
        'get': exceptions.GETRequestError,
        'post': exceptions.POSTRequestError,
        'put': exceptions.PUTRequestError,
        'delete': exceptions.DELETERequestError,
        'other': exceptions.APIRequestError
    }
    if type(api_response) == dict:
        error_found, error_details = get_error_from_json(api_response, v1=True, fail_on_no_results=fail_on_no_results)
    elif api_response.status_code != 200:
        if type(api_response.text) == str and api_response.text.startswith('<html>'):
            error_msg = get_error_from_html(api_response.text, v1=True)
            raise exception_classes.get(query_type)(error_msg)
        else:
            raise exception_classes.get(query_type)
    elif type(api_response.text) == str and api_response.text.startswith('<response'):
        error_found, error_details = get_error_from_xml(api_response.text, endpoint, fail_on_no_results)
    else:
        error_found, error_details = False, None
    if error_found:
        error_code, error_msg, exc_msg = error_details
        if error_code == 404 and fail_on_no_results:
            raise exceptions.NotFoundResponseError(exc_msg)
        else:
            raise exception_classes.get(query_type)(exc_msg)



# -*- coding: utf-8 -*-
"""
:Package:        khoros.utils
:Synopsis:       This is the ``__init__`` module for the khoros.utils modules
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  20 Feb 2020
"""

__all__ = ['core_utils', 'helper', 'version']



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.log_utils
:Synopsis:          Collection of logging utilities and functions
:Usage:             ``from khoros.utils import log_utils``
:Example:           ``logger = log_utils.initialize_logging(__name__)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     23 May 2022
"""

import os
import sys
import logging
import logging.handlers
from pathlib import Path

LOGGING_DEFAULTS = {
    'logger_name': __name__,
    'log_level': 'info',
    'formatter': logging.Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s'),
    'date_format': '%Y-%m-%d %I:%M:%S',
}
HANDLER_DEFAULTS = {
    'file_log_level': 'info',
    'console_log_level': 'warning',
    'syslog_log_level': 'info',
    'syslog_address': 'localhost',
    'syslog_port': 514,
}


def initialize_logging(logger_name=None, log_level=None, formatter=None, debug=None, no_output=None, file_output=None,
                       file_log_level=None, log_file=None, overwrite_log_files=None, console_output=None,
                       console_log_level=None, syslog_output=None, syslog_log_level=None, syslog_address=None,
                       syslog_port=None):
    """This function initializes logging for the khoros library."""
    # TODO: Complete the docstring above with parameters
    logger_name, log_levels, formatter = _apply_defaults(logger_name, formatter, debug, log_level, file_log_level,
                                                         console_log_level, syslog_log_level)
    log_level, file_log_level, console_log_level, syslog_log_level = _get_log_levels_from_dict(log_levels)
    logger = logging.getLogger(logger_name)
    logger = _set_logging_level(logger, log_level)
    logger = _add_handlers(logger, formatter, no_output, file_output, file_log_level, log_file, overwrite_log_files,
                           console_output, console_log_level, syslog_output, syslog_log_level, syslog_address,
                           syslog_port)
    return logger


class LessThanFilter(logging.Filter):
    """This class allows filters to be set to limit log levels to only less than a specified level.

    .. versionadded:: 3.0.0

    .. seealso:: `Zoey Greer <https://stackoverflow.com/users/5124424/zoey-greer>`_ is the original author of
                 this class which was provided on `Stack Overflow <https://stackoverflow.com/a/31459386>`_.
    """
    def __init__(self, exclusive_maximum, name=""):
        """This method instantiates the :py:class:`khoros.utils.log_utils.LessThanFilter` class object.

        .. versionadded:: 3.0.0
        """
        super(LessThanFilter, self).__init__(name)
        self.max_level = exclusive_maximum

    def filter(self, record):
        """This method returns a Boolean integer value indicating whether or not a message should be logged.

        .. versionadded:: 3.0.0

        .. note:: A non-zero return indicates that the message will be logged.
        """
        return 1 if record.levelno < self.max_level else 0


def _apply_defaults(_logger_name, _formatter, _debug, _log_level, _file_level, _console_level, _syslog_level):
    """This function applies default values to the configuration settings if not explicitly defined.

    .. versionadded:: 3.0.0

    :param _logger_name: The name of the logger instance
    :type _logger_name: str, None
    :param _formatter: The log format to utilize for the logger instance
    :type _formatter: str, None
    :param _debug: Defines if debug mode is enabled
    :type _debug: bool, None
    :param _log_level: The general logging level for the logger instance
    :type _log_level: str, None
    :returns: The values that will be used for the configuration settings
    """
    _log_levels = {
        'general': _log_level,
        'file': _file_level,
        'console': _console_level,
        'syslog': _syslog_level,
    }
    _logger_name = LOGGING_DEFAULTS.get('logger_name') if not _logger_name else _logger_name
    if _debug:
        for _log_type in _log_levels:
            _log_levels[_log_type] = 'debug'
    else:
        if _log_level:
            for _lvl_type, _lvl_value in _log_levels.items():
                if _lvl_type != 'general' and _lvl_value is None:
                    _log_levels[_lvl_type] = _log_level
        else:
            _log_level = LOGGING_DEFAULTS.get('log_level')
    if _formatter and isinstance(_formatter, str):
        _formatter = logging.Formatter(_formatter)
    _formatter = LOGGING_DEFAULTS.get('formatter') if not _formatter else _formatter
    return _logger_name, _log_levels, _formatter


def _get_log_levels_from_dict(_log_levels):
    """This function returns the individual log level values from a dictionary.

    .. versionadded:: 3.0.0

    :param _log_levels: Dictionary containing log levels for different handlers
    :type _log_levels: dict
    :returns: Individual string values for each handler
    """
    _general = _log_levels.get('general')
    _file = _log_levels.get('file')
    _console = _log_levels.get('console')
    _syslog = _log_levels.get('syslog')
    return _general, _file, _console, _syslog


def _set_logging_level(_logger, _log_level):
    """This function sets the logging level for a :py:class:`logging.Logger` instance.

    .. versionadded:: 3.0.0

    :param _logger: The :py:class:`logging.Logger` instance
    :type _logger: Logger
    :param _log_level: The log level as a string (``debug``, ``info``, ``warning``, ``error`` or ``critical``)
    :type _log_level: str
    :returns: The :py:class:`logging.Logger` instance with a logging level set where applicable
    """
    if _log_level == 'debug':
        _logger.setLevel(logging.DEBUG)
    elif _log_level == 'info':
        _logger.setLevel(logging.INFO)
    elif _log_level == 'warning':
        _logger.setLevel(logging.WARNING)
    elif _log_level == 'error':
        _logger.setLevel(logging.ERROR)
    elif _log_level == 'critical':
        _logger.setLevel(logging.CRITICAL)
    return _logger


def _add_handlers(_logger, _formatter, _no_output, _file_output, _file_log_level, _log_file, _overwrite_log_files,
                  _console_output, _console_log_level, _syslog_output, _syslog_log_level, _syslog_address,
                  _syslog_port):
    # TODO: Add docstring
    if _no_output or not any((_file_output, _console_output, _syslog_output)):
        _logger.addHandler(logging.NullHandler())
    else:
        if _file_output:
            # Add the FileHandler to the Logger object
            _logger = _add_file_handler(_logger, _file_log_level, _log_file, _overwrite_log_files, _formatter)
        if _console_output:
            # Add the StreamHandler to the Logger object
            _logger = _add_stream_handler(_logger, _console_log_level, _formatter)
        if _syslog_output:
            # Add the SyslogHandler to the Logger object
            _logger = _add_syslog_handler(_logger, _syslog_log_level, _formatter, _syslog_address, _syslog_port)
    return _logger


def _add_file_handler(_logger, _log_level, _log_file, _overwrite, _formatter):
    """This function adds a :py:class:`logging.FileHandler` to the :py:class:`logging.Logger` instance.

    .. versionadded:: 3.0.0

    :param _logger: The :py:class:`logging.Logger` instance
    :type _logger: Logger
    :param _log_level: The log level to set for the handler
    :type _log_level: str
    :param _log_file: The log file (as a file name or a file path) to which messages should be written

    .. note:: If a file path isn't provided then the default directory is the home directory of the user instantiating
              the :py:class:`logging.Logger` object. If a file name is also no provided then it will default to
              using ``khoros.log`` as the file name.

    :param _overwrite: Determines if messages should be appended to the file (default) or overwrite it
    :type _overwrite: bool
    :param _formatter: The :py:class:`logging.Formatter` to apply to messages passed through the handler
    :type _formatter: Formatter
    :returns: The :py:class:`logging.Logger` instance with the added :py:class:`logging.FileHandler`
    """
    # Define the log file to use
    _home_dir = str(Path.home())
    if _log_file:
        if not any((('/' in _log_file), ('\\' in _log_file))):
            _log_file = os.path.join(_home_dir, _log_file)
    else:
        _log_file = os.path.join(_home_dir, 'khoros.log')

    # Identify if log file should be overwritten
    _write_mode = 'w' if _overwrite else 'a'

    # Instantiate the handler
    _handler = logging.FileHandler(_log_file, _write_mode)
    _log_level = HANDLER_DEFAULTS.get('file_log_level') if not _log_level else _log_level
    _handler = _set_logging_level(_handler, _log_level)
    _handler.setFormatter(_formatter)

    # Add the handler to the logger
    _logger.addHandler(_handler)
    return _logger


def _add_stream_handler(_logger, _log_level, _formatter):
    """This function adds a :py:class:`logging.StreamHandler` to the :py:class:`logging.Logger` instance.

    .. versionadded:: 3.0.0

    :param _logger: The :py:class:`logging.Logger` instance
    :type _logger: Logger
    :param _log_level: The log level to set for the handler
    :type _log_level: str
    :param _formatter: The :py:class:`logging.Formatter` to apply to messages passed through the handler
    :type _formatter: Formatter
    :returns: The :py:class:`logging.Logger` instance with the added :py:class:`logging.StreamHandler`
    """
    _log_level = HANDLER_DEFAULTS.get('console_log_level') if not _log_level else _log_level
    _stdout_levels = ['DEBUG', 'INFO']
    if _log_level.upper() in _stdout_levels:
        _logger = _add_split_stream_handlers(_logger, _log_level, _formatter)
    else:
        _handler = logging.StreamHandler()
        _handler = _set_logging_level(_handler, _log_level)
        _handler.setFormatter(_formatter)
        _logger.addHandler(_handler)
    return _logger


def _add_split_stream_handlers(_logger, _log_level, _formatter):
    """This function splits messages into q ``stdout`` or ``stderr`` handler depending on the log level.

    .. versionadded:: 3.0.0

    .. seealso:: Refer to the documentation for the :py:class:`khoros.utils.log_utils.LessThanFilter` for
                 more information on how this filtering is implemented and for credit to the original author.

    :param _logger: The :py:class:`logging.Logger` instance
    :type _logger: Logger
    :param _log_level: The log level provided for the stream handler (i.e. console output)
    :type _log_level: str
    :param _formatter: The :py:class:`logging.Formatter` to apply to messages passed through the handlers
    :type _formatter: Formatter
    :returns: The logger instance with the two handlers added
    """
    # Configure and add the STDOUT handler
    _stdout_handler = logging.StreamHandler(sys.stdout)
    _stdout_handler = _set_logging_level(_stdout_handler, _log_level)
    _stdout_handler.addFilter(LessThanFilter(logging.WARNING))
    _stdout_handler.setFormatter(_formatter)
    _logger.addHandler(_stdout_handler)

    # Configure and add the STDERR handler
    _stderr_handler = logging.StreamHandler(sys.stderr)
    _stderr_handler.setLevel(logging.WARNING)
    _stderr_handler.setFormatter(_formatter)
    _logger.addHandler(_stderr_handler)

    # Return the logger with the added handlers
    return _logger


def _add_syslog_handler(_logger, _log_level, _formatter, _address, _port):
    # TODO: Add docstring
    _log_level = HANDLER_DEFAULTS.get('syslog_log_level') if not _log_level else _log_level
    _address = HANDLER_DEFAULTS.get('syslog_address') if not _address else _address
    _port = HANDLER_DEFAULTS.get('syslog_port') if not _port else _port
    _handler = logging.handlers.SysLogHandler(address=(_address, _port))
    _handler = _set_logging_level(_handler, _log_level)
    _handler.setFormatter(_formatter)
    _logger.addHandler(_handler)
    return _logger



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.tests.test_studio
:Synopsis:          This module is used by pytest to verify that the ``studio`` module functions properly
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     03 Oct 2022
"""

from . import resources


def test_studio_functions():
    """This function tests the various Studio-related functions.

    .. versionadded:: 5.1.2
    """
    # Instantiate the core object
    khoros_object = resources.get_core_object()

    # Test if SDK is installed
    sdk_installed = khoros_object.studio.sdk_installed()
    assert isinstance(sdk_installed, bool)

    # Test if Node.js is installed
    node_installed = khoros_object.studio.node_installed()
    assert isinstance(node_installed, bool)

    # Tset if npm is installed
    npm_installed = khoros_object.studio.npm_installed()
    assert isinstance(npm_installed, bool)

    # Test retrieving the Node.js version
    node_version = khoros_object.studio.get_node_version()
    assert node_version is None or isinstance(node_version, str)

    # Test retrieving the npm version
    npm_version = khoros_object.studio.get_npm_version()
    assert npm_version is None or isinstance(npm_version, str)

    # Test retrieving the SDK version
    sdk_version = khoros_object.studio.get_sdk_version()
    assert sdk_version is None or isinstance(sdk_version, str)



# -*- coding: utf-8 -*-
"""
:Module:            khoros.structures.nodes
:Synopsis:          This module contains functions specific to nodes within the Khoros Community platform
:Usage:             ``from khoros.structures import nodes``
:Example:           ``node_id = nodes.get_id(url)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     10 Mar 2021
"""

import re

from . import base
from .. import liql, errors
from ..utils import log_utils

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)


def get_node_id(url=None, node_type=None, node_details=None):
    """This function retrieves the Node ID for a given node within a URL.

    :param url: The URL from which to parse out the Node ID
    :type url: str, None
    :param node_type: The node type (e.g. ``blog``, ``tkb``, ``message``, etc.) for the object in the URL
    :type node_type: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The Node ID retrieved from the URL
    :raises: :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.NodeIDNotFoundError`,
             :py:exc:`khoros.errors.exceptions.NodeTypeNotFoundError`
    """
    if not url and not node_details:
        raise errors.exceptions.MissingRequiredDataError("A node URL or dictionary with node details must be given.")
    if node_details:
        node_id = node_details['id'].split(':')[1]
    else:
        if not node_type:
            # Attempt to get the Node Type from the URL
            node_type = get_node_type_from_url(url)
        elif node_type not in Mapping.node_url_mapping:
            node_type = _get_node_type_identifier(node_type)
        node_url_segment = Mapping.node_url_mapping.get(node_type) + '/'
        if node_url_segment not in url:
            raise errors.exceptions.InvalidNodeTypeError(val=node_type)
        node_id = re.sub(r'/.*$', '', re.sub(r'^.*' + node_url_segment, '', url))
        if not node_id or len(node_id) == 0:
            raise errors.exceptions.NodeIDNotFoundError(val=url)
    return node_id


def _get_node_type_identifier(_node_type_lookup):
    """This function attempts to identify the appropriate node type for a function.

    :param _node_type_lookup: The value to look up as a node type
    :type _node_type_lookup: str
    :returns: The appropriate node type (if found)
    :raises: :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError`
    """
    if _node_type_lookup in Mapping.proper_name_mapping:
        _node_type = Mapping.proper_name_mapping.get(_node_type_lookup)
    else:
        raise errors.exceptions.InvalidNodeTypeError(val=_node_type_lookup)
    return _node_type


def get_node_type_from_url(url):
    """This function attempts to retrieve a node type by analyzing a supplied URL.

    :param url: The URL from which to extract the node type
    :type url: str
    :returns: The node type based on the URL provided
    :raises: :py:exc:`khoros.errors.exceptions.NodeTypeNotFoundError`
    """
    node_type = None
    for node_type_name, node_url_code in Mapping.node_url_mapping.items():
        if node_url_code in url:
            node_type = node_type_name
            break
    if not node_type:
        raise errors.exceptions.NodeTypeNotFoundError(val=url)
    return node_type


def get_total_node_count(khoros_object):
    """This function returns the total number of nodes within the Khoros Community environment.

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :returns: The total number of nodes as an integer
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return liql.get_total_count(khoros_object, 'nodes')


def node_exists(khoros_object, node_id=None, node_url=None):
    """This function checks to see if a node exists.

    .. versionadded:: 2.7.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param node_id: The ID of the node to check
    :type node_id: str, None
    :param node_url: The URL of the node to check
    :type node_url: str, None
    :returns: Boolean value indicating whether or not the node already exists
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return base.structure_exists(khoros_object, 'node', node_id, node_url)


def get_node_details(khoros_object, identifier, first_item=True):
    """This function returns a dictionary of node configuration settings.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str
    :param first_item: Filters the response data to the first item returned (``True`` by default)
    :type first_item: bool
    :returns: The node details within a dictionary
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return base.get_details(khoros_object, identifier, 'node', first_item)


def get_node_field(khoros_object, field, identifier=None, node_details=None):
    """This function returns a specific node field from the Khoros Community API.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param field: The field from the :py:class:`khoros.structures.base.Mapping` class whose value should be returned
    :type field: str
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The requested field in its native format
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if field == 'view_href' or field == 'url':
        field_value = get_url(khoros_object, identifier, node_details)
    elif field == 'id':
        field_value = get_node_id(identifier, node_details=node_details)
    else:
        field_value = base.get_structure_field(khoros_object, field, identifier,
                                               structure_type='node', details=node_details)
    return field_value


def get_url(khoros_object, node_id=None, node_details=None):
    """This function returns the full URL of a given Node ID.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param node_id: The Node ID with which to identify the node
    :type node_id: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The node URl as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if not node_id and not node_details:
        raise errors.exceptions.MissingRequiredDataError("A Node ID or a dictionary with node details must be given.")
    elif not node_id or ('/' in node_id and node_details):
        node_id = get_node_id(node_details=node_details)
    query = f"SELECT view_href FROM nodes WHERE id = '{node_id}'"       # nosec
    response = liql.perform_query(khoros_object, liql_query=query)
    return response['data']['items'][0]['view_href']


def get_type(khoros_object, identifier, node_details=None):
    """This function returns the full URL of a given Node ID.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The node URl as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return get_node_field(khoros_object, 'node_type', identifier, node_details)


def get_discussion_style(khoros_object, identifier, node_details=None):
    """This function returns the full URL of a given Node ID.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The node URl as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return get_node_field(khoros_object, 'discussion_style', identifier, node_details)


def get_title(khoros_object, identifier=None, full_title=True, short_title=False, node_details=None):
    """This function retrieves the full and/or short title of the node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param full_title: Determines if the full title of the node should be returned (``True`` by default)
    :type full_title: bool
    :param short_title: Determines if the short title of the node should be returned (``False`` by default)
    :type short_title: bool
    :param node_details: Dictionary containing node details (optional)
    :type node_details: dict, None
    :returns: The node title(s) as a string or a tuple of strings
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if not full_title and not short_title:
        exc_msg = "Must return at least the full title or the short title."
        raise errors.exceptions.MissingRequiredDataError(exc_msg)
    if not node_details:
        node_details = get_node_details(khoros_object, identifier)
    titles = (node_details['title'], node_details['short_title'])
    if not short_title:
        titles = titles[0]
    elif not full_title:
        titles = titles[1]
    return titles


def get_description(khoros_object, identifier, node_details=None):
    """This function returns the description of a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The node description as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return get_node_field(khoros_object, 'description', identifier, node_details)


def get_parent_type(khoros_object, identifier, node_details=None):
    """This function returns the parent type of a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The parent type as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return get_node_field(khoros_object, 'parent_type', identifier, node_details)


def get_parent_id(khoros_object, identifier, node_details=None, include_prefix=False):
    """This function returns the Parent ID of a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :param include_prefix: Determines if the prefix (e.g. ``category:``) should be included (``False`` by default)
    :type include_prefix: bool
    :returns: The Parent ID as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    parent_id = get_node_field(khoros_object, 'parent_id', identifier, node_details)
    if not include_prefix:
        parent_id = parent_id.split(':')[1]
    return parent_id


def get_parent_url(khoros_object, identifier, node_details=None):
    """This function returns the parent URL of a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The parent URL as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    parent_id = get_parent_id(khoros_object, identifier, node_details, include_prefix=False)
    return get_url(khoros_object, parent_id)


def get_root_type(khoros_object, identifier, node_details=None):
    """This function returns the root category type of a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The root category type as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return get_node_field(khoros_object, 'root_type', identifier, node_details)


def get_root_id(khoros_object, identifier, node_details=None, include_prefix=False):
    """This function returns the Root Category ID of a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :param include_prefix: Determines if the prefix (e.g. ``category:``) should be included (``False`` by default)
    :type include_prefix: bool
    :returns: The Root Category ID as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    root_id = get_node_field(khoros_object, 'root_id', identifier, node_details)
    if not include_prefix:
        root_id = root_id.split(':')[1]
    return root_id


def get_root_url(khoros_object, identifier, node_details=None):
    """This function returns the root category URL of a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The root category URL as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    root_id = get_root_id(khoros_object, identifier, node_details, include_prefix=False)
    return get_url(khoros_object, root_id)


# noinspection PyTypeChecker
def get_avatar_url(khoros_object, identifier, node_details=None, original=True, tiny=False, small=False,
                   medium=False, large=False):
    """This function retrieves one or more avatar URLs for a given node.

            .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :param original: Indicates if the URL for the image in its original size should be returned (``True`` by default)
    :type original: bool
    :param tiny: Indicates if the URL for the image in a tiny size should be returned (``False`` by default)
    :type tiny: bool
    :param small: Indicates if the URL for the image in a small size should be returned (``False`` by default)
    :type small: bool
    :param medium: Indicates if the URL for the image in a medium size should be returned (``False`` by default)
    :type medium: bool
    :param large: Indicates if the URL for the image in a large size should be returned (``False`` by default)
    :type large: bool
    :returns: A single URL as a string (default) or a dictionary of multiple URLs by size
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if not original and not tiny and not small and not medium and not large:
        raise errors.exceptions.MissingRequiredDataError("At least one file size must be enabled.")
    avatar_details = get_node_field(khoros_object, 'avatar_details', identifier, node_details)
    avatar_urls, file_sizes = {}, {'tiny': tiny, 'small': small, 'medium': medium, 'large': large}
    if original:
        avatar_urls['original'] = avatar_details['tiny_href'].split('/image-size/')[0]
    for size in file_sizes:
        if file_sizes.get(size):
            avatar_urls[size] = avatar_details[Mapping.avatar_size_mapping.get(size)]
    if len(avatar_urls) == 1:
        avatar_urls = list(avatar_urls.values())[0]
    return avatar_urls


def get_creation_date(khoros_object, identifier, node_details=None, friendly=False):
    """This function returns the creation date of a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :param friendly: Determines whether to return the "friendly" date (e.g. ``Friday``) instead (``False`` by default)
    :type friendly: bool
    :returns: The creation date as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    # TODO: Allow a format to be specified and the ability to parse as a datetime object if needed
    if friendly:
        creation_date = get_node_field(khoros_object, 'creation_date_friendly', identifier, node_details)
    else:
        creation_date = get_node_field(khoros_object, 'creation_date', identifier, node_details)
    return creation_date


def get_depth(khoros_object, identifier, node_details=None):
    """This function returns the depth of a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The depth as an integer
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return get_node_field(khoros_object, 'depth', identifier, node_details)


def get_position(khoros_object, identifier, node_details=None):
    """This function returns the position of a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The position as an integer
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return get_node_field(khoros_object, 'position', identifier, node_details)


def is_hidden(khoros_object, identifier, node_details=None):
    """This function identifies whether or not a given node is hidden.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: Boolean indicating whether or not the node is hidden
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return get_node_field(khoros_object, 'hidden', identifier, node_details)


def get_views(khoros_object, identifier, node_details=None):
    """This function returns the views for a given node.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The Node ID or Node URL with which to identify the node
    :type identifier: str, None
    :param node_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type node_details: dict, None
    :returns: The views as an integer
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return get_node_field(khoros_object, 'views', identifier, node_details)


class Mapping:
    """This class includes dictionaries that map API fields and values to those used in this SDK."""
    node_url_mapping = {
        'category': 'ct-p',
        'blog': 'bg-p',
        'contest': 'con-p',
        'board': 'bd-p',
        'group': 'gp-p',
        'idea': 'idb-p',
        'message': 'm-p',
        'qa': 'qa-p',
        'tkb': 'tkb-p'
    }
    proper_name_mapping = {
        'Category': 'category',
        'Blog': 'blog',
        'Board': 'board',
        'Contest': 'contest',
        'Forum': 'forum',
        'Group': 'group',
        'Idea Exchange': 'idea',
        'Message': 'message',
        'Q&A': 'qa',
        'TKB': 'tkb'
    }
    avatar_size_mapping = {
        'tiny': 'tiny_href',
        'small': 'small_href',
        'medium': 'medium_href',
        'large': 'large_href'
    }



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.core_utils
:Synopsis:          Collection of supporting utilities and functions to complement the primary modules
:Usage:             ``from khoros.utils import core_utils``
:Example:           ``encoded_string = core_utils.encode_url(decoded_string)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     23 May 2022
"""

import os
import base64
import random
import string
import warnings
import subprocess
import urllib.parse
from html import unescape

from .. import errors
from . import log_utils

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)


def url_encode(raw_string):
    """This function encodes a string for use in URLs.

    :param raw_string: The raw string to be encoded
    :type raw_string: str
    :returns: The encoded string
    """
    return urllib.parse.quote_plus(raw_string)


def url_decode(encoded_string):
    """This function decodes a url-encoded string.

    :param encoded_string: The url-encoded string
    :type encoded_string: str
    :returns: The unencoded string
    """
    return urllib.parse.unquote_plus(encoded_string)


def decode_html_entities(html_string):
    """This function converts HTML entities (e.g. ``&amp;``, ``&apos;``, etc.) back to their original characters.

    :param html_string: The string containing HTML entities to be decoded
    :type html_string: str
    :returns: The string with decoded HTML entities
    """
    return unescape(html_string)


def decode_binary(binary):
    """This function decodes a binary into a UTF-8 encoded string.

    .. versionadded:: 2.6.0

    :param binary: The binary to be decoded
    :returns: The properly decoded string
    :raises: :py:exc:`TypeError`, :py:exc:`ValueError`
    """
    return binary.decode('utf-8')


def encode_base64(object_to_encode, str_encoding='utf-8', url_encode_object=False, return_bytes=False):
    """This function encodes a string or bytes-like object

    .. versionadded:: 3.0.0

    :param object_to_encode: The string or bytes-like object to encode as base64
    :param str_encoding: Defines the encoding (``utf-8`` by default) to utilize
    :type str_encoding: str
    :param url_encode_object: Determines if the base64 string should be url-encoded (``False`` by default)
    :type url_encode_object: bool
    :param return_bytes: Determines if the base64-encoded object should be returned as a bytes-like object rather
                         than a string (``False`` by default)
    :returns: The encoded object as a string or bytes-like object
    :raises: :py:exc:`TypeError`
    """
    if isinstance(object_to_encode, str):
        object_to_encode = object_to_encode.encode(str_encoding)
    base64_object = base64.b64encode(object_to_encode)
    if not return_bytes:
        base64_object = base64_object.decode(str_encoding)
        if url_encode_object:
            base64_object = url_encode(base64_object)
    return base64_object


def run_cmd(cmd, return_type='dict', shell=False, decode_output=True, strip_output=False,
            exclude_stdout=False, exclude_stderr=False, exclude_return_code=False):
    """This function executes a shell command on the operating system.

    .. versionchanged:: 3.5.0
       The default value of the ``shell`` parameter has been changed to ``False`` to avoid unnecessary
       `security <https://bandit.readthedocs.io/en/latest/plugins/b602_subprocess_popen_with_shell_equals_true.html>`_
       risk and added a logged warning if the value is manually set to ``True``.

    .. versionadded:: 2.5.1

    :param cmd: The command to be executed
    :type cmd: str
    :param return_type: Determines the format in which the results should be returned (``dict`` by default)
    :type return_type: str
    :param shell: Determines if the ``shell`` argument in the :py:func:`subprocess.run` function should be ``True``
    :type shell: bool
    :param decode_output: Determines if the binary output should be decoded as a UTF-8 string (``True`` by default)
    :type decode_output: bool
    :param strip_output: Determines if the escape character(s) should be stripped from the output (``False`` by default)
    :type strip_output: bool
    :param exclude_stdout: Determines if the ``stdout`` output should be excluded (``False`` by default)
    :type exclude_stdout: bool
    :param exclude_stderr: Determines if the ``stderr`` output should be excluded (``False`` by default)
    :type exclude_stderr: bool
    :param exclude_return_code: Determines if the return code from the command should be excluded (``False`` by default)
    :type exclude_return_code: bool
    :returns: The results from the executed script
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if exclude_stdout and exclude_stderr and exclude_return_code:
        raise errors.exceptions.MissingRequiredDataError("At least one output type must be enabled.")
    if shell:
        warn_msg = "It is recommended that the shell parameter be set to False to avoid introducing risk of " \
                   "shell injection attacks in your code."
        logger.warning(warn_msg)
    output = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell)
    stdout, stderr, return_code = output.stdout, output.stderr, output.returncode
    results = {
        'stdout': stdout,
        'stderr': stderr,
        'return_code': return_code
    }
    for stream in ('stdout', 'stderr'):
        if decode_output:
            results[stream] = decode_binary(results.get(stream))
        if strip_output:
            results[stream] = results.get(stream).strip()
    output_types = {'stdout': exclude_stdout, 'stderr': exclude_stderr, 'return_code': exclude_return_code}
    for output_type, excluded in output_types.items():
        if excluded:
            del results[output_type]
    if return_type == 'list':
        results = list(results.values())
    elif return_type == 'tuple':
        results = tuple(results.values())
    else:
        if return_type != 'dict':
            raise ValueError(f"'{return_type}' is not a valid return type.")
    return results


def _is_zero_length(_element):
    """This function checks to see if an element has a zero length.

    :param _element: The element of which the length will be checked
    :returns: Boolean value stating whether or not the length of the element is zero
    """
    return True if len(_element) == 0 else False


def _structure_query_string(_url_dict, _no_encode):
    """This function constructs a query string where one or more fields must not be URL-encoded.

    :param _url_dict: Dictionary of URL query string keys and values
    :type _url_dict: dict
    :param _no_encode: Designates any dictionary keys (i.e. field names) whose values should not be URL-encoded
    :type _no_encode: list, tuple, set, str, None
    :returns: The URL query string in string format
    """
    if type(_no_encode) == str:
        _no_encode = (_no_encode, )
    _delimiters = {True: "", False: "&"}
    _query_string = ""
    for _field_name, _field_value in _url_dict.items():
        if _field_name not in _no_encode:
            _field_value = url_encode(_field_value)
        _delimiter = _delimiters.get(_is_zero_length(_query_string))
        _query_string = f"{_query_string}{_delimiter}{_field_name}={_field_value}"
    return _query_string


def encode_query_string(url_dict, no_encode=None, json_payload=False):
    """This function compiles a URL query string from a dictionary of parameters.

    .. versionchanged:: 3.2.0
       Introduced the ability to pass the query parameters as JSON payload to avoid URI length limits.

    :param url_dict: Dictionary of URL query string keys and values
    :type url_dict: dict
    :param no_encode: Designates any dictionary keys (i.e. field names) whose values should not be URL-encoded
    :type no_encode: list, tuple, set, str, None
    :param json_payload: Determines if query parameters should be passed as JSON payload rather than in the URI
                         (``False`` by default)
    :type json_payload: bool
    :returns: The URL query string in string format
    """
    if json_payload:
        # Structure the query string using only the field names
        query_string = ""
        for field in url_dict.keys():
            delimiter = "&" if len(query_string) > 0 else ""
            query_string = f"{query_string}{delimiter}{field}"
    elif no_encode:
        query_string = _structure_query_string(url_dict, no_encode)
    else:
        query_string = urllib.parse.urlencode(url_dict)
    return query_string


def is_numeric(value):
    """This function checks whether or not a value is numeric either as an integer or a numeric string.

    .. versionadded:: 2.3.0

    :param value: The value to be examined
    :type value: str, int
    :returns: Boolean value indicating if the examined value is numeric
    """
    return True if type(value) == int or (type(value) == str and value.isnumeric()) else False


def convert_set(iterable, convert_to='list'):
    """This function casts a ``set`` variable to be a ``list`` instead so that it can be scriptable.

    :param iterable: The iterable to be evaluated to see if it has a ``set`` type
    :param convert_to: Defines if the iterable should be cast to a ``list`` (default) or a ``tuple``
    :type convert_to: str
    :returns: The converted variable as a ``list`` or ``tuple`` (or untouched if not a ``set``)
    """
    if type(iterable) == set:
        if convert_to == 'tuple':
            iterable = tuple(iterable)
        else:
            iterable = list(iterable)
    return iterable


def convert_single_value_to_tuple(value):
    """This function converts a single value of nearly any type into a tuple.

    .. versionchanged:: 3.2.0
       The function has been aesthetically updated to be more PEP8 compliant.

    .. versionadded:: 2.3.0

    :param value: The value to convert into a tuple
    """
    value = (value,)
    return value


def convert_string_to_tuple(value, delimiter=''):
    """THis function converts a value to a tuple if in string format.

    .. versionchanged:: 3.5.0
       The typecheck has been updated to use ``isinstance`` and the function can now split delimited strings as needed.

    .. versionadded:: 2.3.0

    :param value: The potential string to convert
    :type value: str
    :param delimiter: The value (e.g. ``,``) used to separate values in a delimited string (empty by default)
    :returns: The tuple (if original value was in string format) or the original value/type
    """
    if isinstance(value, str):
        if delimiter and delimiter in value:
            value = value.split(delimiter)
        else:
            value = convert_single_value_to_tuple(value)
    return value


def is_iterable(var):
    """This function identifies if a given variable is an iterable.

    .. versionadded:: 3.5.0

    :param var: The variable to check
    :returns: A boolean value indicating whether or not the variable is an iterable
    """
    is_iter = any((isinstance(var, list), isinstance(var, tuple), isinstance(var, set),
                   isinstance(var, type({}.keys())), isinstance(var, type({}.values()))))
    return is_iter


def get_random_string(length=32, prefix_string=""):
    """This function returns a random alphanumeric string to use as a salt or password.

    :param length: The length of the string (``32`` by default)
    :type length: int
    :param prefix_string: A string to which the salt should be appended (optional)
    :type prefix_string: str
    :returns: The alphanumeric string
    """
    return f"{prefix_string}{''.join([random.choice(string.ascii_letters + string.digits) for _ in range(length)])}"


def display_warning(warn_msg):
    """This function displays a :py:exc:`UserWarning` message via the :py:mod:`warnings` module.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    .. versionadded:: 2.1.0

    :param warn_msg: The message to be displayed
    :type warn_msg: str
    :returns: None
    """
    warnings.warn(warn_msg, UserWarning)


def get_file_type(file_path):
    """This function attempts to identify if a given file path is for a YAML or JSON file.

    .. versionadded:: 2.2.0

    :param file_path: The full path to the file
    :type file_path: str
    :returns: The file type in string format (e.g. ``yaml`` or ``json``)
    :raises: :py:exc:`FileNotFoundError`, :py:exc:`khoros.errors.exceptions.UnknownFileTypeError`
    """
    file_type = 'unknown'
    if os.path.isfile(file_path):
        if file_path.endswith('.json'):
            file_type = 'json'
        elif file_path.endswith('.yml') or file_path.endswith('.yaml'):
            file_type = 'yaml'
        else:
            display_warning(f"Unable to recognize the file type of '{file_path}' by its extension.")
            with open(file_path) as cfg_file:
                for line in cfg_file:
                    if line.startswith('#'):
                        continue
                    else:
                        if '{' in line:
                            file_type = 'json'
                            break
        if file_type == 'unknown':
            raise errors.exceptions.UnknownFileTypeError(file=file_path)
    else:
        raise FileNotFoundError(f"Unable to locate the following file: {file_path}")
    return file_type


def convert_dict_id_values_to_strings(dict_list):
    """This function ensures that the ``id`` keys in a list of dictionaries use string values.

    :param dict_list: List (or tuple) of dictionaries (or a single dictionary) containing API object data
    :type dict_list: list, tuple, dict, None
    :returns: A new dictionary list with properly formatted ``id`` values
    :raises: :py:exc:`TypeError`
    """
    dict_list = [dict_list] if isinstance(dict_list, dict) else dict_list
    new_dict_list = []
    for single_dict in dict_list:
        if not isinstance(single_dict, dict):
            raise TypeError("The 'dict_list' argument must be a dictionary or a list of dictionaries.")
        if 'id' in single_dict and not isinstance(single_dict.get('id'), str):
            single_dict['id'] = str(single_dict.get('id'))
        new_dict_list.append(single_dict)
    return new_dict_list


def convert_dict_list_to_simple_list(dict_list, fields):
    """This function converts a list of dictionaries into a simple list consisting of the provided field(s).

    .. versionadded:: 3.5.0

    :param dict_list: The original list of dictionaries
    :type dict_list: list
    :param fields: The field(s) with which to filter the dictionary list into a simple list
    :type fields: str, tuple, list
    :returns: The simple list of stings or tuples depending on the number of fields
    """
    new_list = []
    fields = convert_string_to_tuple(fields, ',')
    for field_dict in dict_list:
        field_list = []
        for field in fields:
            if field_dict.get(field):
                field_list.append(field_dict.get(field))
        if not field_list:
            field_list[0] = ''
        new_field = field_list[0] if len(field_list) == 1 else tuple(field_list)
        new_list.append(new_field)
    return new_list


def convert_list_values(values_list, convert_to='str', split_values=False, split_delimiter=','):
    """This function converts the values in a list to a different type.

    :param values_list: The list of values to be converted
    :type values_list: list, tuple, set
    :param convert_to: One of the following types: ``str`` (Default), ``int``, ``float``, ``tuple`` or ``set``
    :param split_values: Determines if the values should be split with a specific delimiter (``False`` by default)

                         .. note:: This only applies when converting to the ``tuple`` or ``set`` types.

    :type split_values: bool
    :param split_delimiter: The delimiter for which to split the values when applicable (comma by default)
    :type split_delimiter: str
    :returns: A new list of converted values
    :raises: :py:exc:`TypeError`, :py:exc:`ValueError`
    """
    new_list = []
    for value in values_list:
        if convert_to == 'str':
            new_list.append(str(value))
        elif convert_to == 'int':
            new_list.append(int(value))
        elif convert_to == 'float':
            new_list.append(float(value))
        elif convert_to == 'tuple':
            value = tuple(value.split(split_delimiter)) if split_values else (value, )
            new_list.append(value)
        elif convert_to == 'set':
            value = set(value.split(split_delimiter)) if split_values else {value}
            new_list.append(value)
    return new_list


def extract_key_values_from_dict_list(key_name, dict_list, exclude_if_present=None, convert_to_string=True):
    """This function extracts values for a specific key from a list of dictionaries.

    :param key_name: The name of the dictionary key from which to extract the value(s)
    :type key_name: str
    :param dict_list: The list of dictionaries (or single dictionary) from which to extract the value(s)
    :type dict_list: list, dict
    :param exclude_if_present: Will skip extracting the key value if this given key is also present (Optional)
    :type exclude_if_present: str, None
    :param convert_to_string: Determines if the values should be converted to string format (``True`` by default)
    :type convert_to_string: bool
    :returns: A list of values extracted from the dictionary list for the given key
    :raises: :py:exc:`TypeError`
    """
    value_list, dict_list = [], [dict_list] if isinstance(dict_list, dict) else dict_list
    for single_dict in dict_list:
        if key_name in single_dict:
            skip_dict = True if exclude_if_present and exclude_if_present in single_dict else False
            if not skip_dict:
                key_value = str(single_dict.get(key_name)) if convert_to_string else single_dict.get(key_name)
                value_list.append(key_value)
    return value_list


def remove_tld(url, strip_anchors=True):
    """This function removes the top-level domain (TLD) from a Khoros Community platform URL.

    :param url: The URL from which the TLD should be removed
    :type url: str
    :param strip_anchors: Determines if anchors (e.g. ``#top``) should be stripped (``True`` by default)
    :type strip_anchors: bool
    :returns: The URL beginning with ``/t5/``
    :raises: :py:exc:`khoros.errors.exceptions.InvalidURLError`
    """
    if '/t5/' not in url:
        raise errors.exceptions.InvalidURLError('The provided URL is not from the Khoros Community platform.')
    url = f"/t5/{url.split('/t5/')[1]}"
    return url.split('#')[0] if strip_anchors and '#' in url else url


def merge_and_dedup(*data):
    """This function merges various data elements into a single, deduplicated list.

    :param data: One or more data elements to merge and deduplicate
    :returns: A merged and deduplicated list of data
    """
    iter_types, unique_list = [list, tuple, set], []
    for element in data:
        if type(element) not in iter_types:
            element = (element,)
        for item in element:
            if item not in unique_list:
                unique_list.append(item)
    return unique_list



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.tests.test_communities
:Synopsis:          This module is used by pytest to verify that the ``communities`` module functions properly
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     29 Sep 2022
"""

import os
import sys

from . import resources

# Define a global variable to define when the package path has been set
package_path_defined = False


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list.

    .. versionadded:: 5.1.0
    """
    global package_path_defined
    if not package_path_defined:
        sys.path.insert(0, os.path.abspath('../..'))
        package_path_defined = True


def test_community_details():
    """This function tests the ability to retrieve community details.

    .. versionchanged:: 5.1.1
       The function has been updated to support GitHub Workflows control data.

    .. versionadded:: 5.1.0
    """
    # Retrieve the control data
    control_data = resources.get_control_data('communities')

    # Instantiate the Khoros object
    set_package_path()
    khoros_object = resources.get_core_object()

    # Test retrieving the community title
    title = khoros_object.communities.get_title()
    assert title == control_data.get('community_title')

    # Test retrieving the community description
    description = khoros_object.communities.get_description()
    assert description == control_data.get('description')

    # Test retrieving the tenant ID
    tenant_id = khoros_object.communities.get_tenant_id()
    assert tenant_id == control_data.get('tenant_id')

    # Test retrieving the primary URL
    primary_url = khoros_object.communities.get_primary_url()
    assert primary_url == control_data.get('primary_url')

    # Test retrieving the sign-out URL
    sign_out_url = khoros_object.communities.get_sign_out_url()
    assert sign_out_url == control_data.get('sign_out_url')

    # Test retrieving the active skin
    active_skin = khoros_object.communities.get_active_skin()
    assert active_skin == control_data.get('active_skin')

    # Test retrieving the community language
    language = khoros_object.communities.get_language()
    assert language == control_data.get('language')

    # Test retrieving the date pattern
    date_pattern = khoros_object.communities.get_date_pattern()
    assert date_pattern == control_data.get('date_pattern')

    # Test retrieving the email confirmation requirement setting
    email_conf_req = khoros_object.communities.email_confirmation_required_to_post()
    assert email_conf_req == control_data.get('email_confirmation')

    # Test retrieving the friendly date setting
    friendly_date = khoros_object.communities.friendly_date_enabled()
    assert friendly_date == control_data.get('friendly_date')

    # Test retrieving the community creation date
    creation_date = khoros_object.communities.get_creation_date()
    assert creation_date == control_data.get('creation_date')

    # Test retrieving the friendly date max age
    friendly_max_age = khoros_object.communities.get_friendly_date_max_age()
    assert friendly_max_age == control_data.get('friendly_max_age')

    # Test retrieving the maximum attachments
    max_attachments = khoros_object.communities.get_max_attachments()
    assert max_attachments == control_data.get('max_attachments')

    # Test retrieving the permitted attachment types
    attachments = khoros_object.communities.get_permitted_attachment_types()
    assert isinstance(attachments, list) and 'jpg' in attachments

    # TODO: Test retrieving Ooyala Player branding ID

    # TODO: Test retrieving top-level breadcrumb setting

    # TODO: Test retrieving show community node in breadcrumb setting

    # TODO: Test retrieving top-level categories enabled setting

    # TODO: Test retrieving top-level categories on community page setting

